(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Translate Hoisted, Closure Converted, Anormalized Ycalculus into Talx86 *)
open Utilities
open Ident
open Variable
open Tal
open Talcon
open Yprim
open Ycon
open Ycalculus
open Format
open Talpp
open Cogen_error
open Cogen_con
open Cogen_util
open Cogen_val
open Cogen_prim
open Cogen_analyze

let rec fold_right3 f args1 args2 args3 base =
  match args1, args2, args3 with
    ([],[],[]) -> base
  | (x1::rest1, x2::rest2, x3::rest3) -> 
      f x1 x2 x3 (fold_right3 f rest1 rest2 rest3 base)
  | _ -> raise (Invalid_argument "fold_right3")

(* Code Generation Assumptions:
   1 All bound identifiers (value and type identifiers) are unique
   2 All functions are closed.  The only free identifiers are labels
      bound in the heap.
   3 All code has been hoisted to the top level. ie: every module
      is a series of lettype definitions followed by a letrec definition
      containing only code followed by an expression containing no
      abstractions.
   4 All code is in anormal form as defined in mil/yanormal with the
     one exception that we allow packs of tuples as right-hand-sides
     as well.
*)
      
(* Code Generation:

   Does not assume constructors are normalized.  Normalizes constructors
   and substitutes all let-bound types during the translation.

   Calling convention:
   Push all arguments on the stack in right to left order.  Then
   push the return type on the stack and jump to the function(ie: use
   the call instruction).  
   esp holds the stack pointer.
   ebp holds a pointer to the current exception handler (allocated on
   stack).
   The result of the function is placed in register eax.
   Exception handlers expect exceptions to be raised with the exception
   in register eax.

   These conventions are revealed by the type translation for functions.
   |(t1,...,tn) -> t| is (fields are initialized by default and 
   sptr == PCstackptr):

   A(p1,p2).{
     esp: sptr(tra :: |t1| :: ... :: |tn| :: p1 @ texn),
     ebp: sptr(texn)
   }
   
   where
    tra = A().{esp: sptr(p1@texn), 
               ebp: sptr(texn), 
               eax: |t|}                       (* return address type *)
    texn = A().{esp: sptr(p2),eax: exn} :: p2  (* exception frame type *)
*)

(* The "heap" is used to add stub code blocks generated in letrecs of 
   recursive tuples.  They can't be added in sequence as translation
   occurs because the code block created is nested inside an instruction
   sequence. 
   
   let heap = [] ref
   let add_heap cb = (heap := cb :: !heap)
*)

(* Find the actual tag number from sum constructors.  
   Not assuming sums are sorted. 
*)
let get_tag css i =
  let rec get_tag css i tagnum vtnum =
    match css with
      [] -> die "cogen_simple_rhs: Ytag: tag larger than sum"
    | []::rest -> 
      	if i = 0 then tagnum 
      	else get_tag rest (i-1) (tagnum+1) vtnum
    | cs::rest ->
        if i = 0 then vtnum
      	else get_tag rest (i-1) tagnum (vtnum+1)
  in
  get_tag css i 0 0
    
(* Generate code for simple right-hand-sides (no nested expressions) 
   Given an cogen environment and a destination register (rdest) for 
   the computation.
   Continue with the instruction list and code blocks of cont.
*)
let rec cogen_simple_rhs 
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env) 
    (rhs : yterm)
    (dest : rhs_finish)
    (cont : ycon * layout -> instruction list * code_blockl list)
    : (instruction list * code_blockl list) =

  let get_dest layout dest =
    match dest with
      Bind_id x ->
	let (r, spill_i, layout') = get_reg layout in
	let layout' = add_id layout x (Rloc r) in
	(r,spill_i,layout')
    | Bind_loc r ->
 	let (spill_i, layout') = spill_reg layout r in
	(r, spill_i, layout')
  in
  (* Mov the value into the destination register rdest and continue *)
  let value_aux env v dest cont =
    let layout = get_layout env in
    let finish () =
      let (r, spill_i, layout') = get_dest layout dest in
      let (v,c) = cogen_val (set_layout env layout') v in
      let (il, cbs) = cont (c,layout') in
      (spill_i @ (Mov (Reg r, v) :: il), cbs)
    in
    match v with 
      Yvar x ->
	(match dest with 
	  Bind_id y ->
	    let (_,_,_,_,deadvars) = layout in
	    if is_dead deadvars y v then
	      (* In this case, we perform copy propagation - no instruction
		 is generated. *)
	      let layout = copyprop_id layout y v in
	      let (_, gamma, _) = get_ctxt env in
	      	cont (Dict.lookup gamma x, layout)
	    else
	      finish ()
	| Bind_loc r ->
	    finish ())
    | _ -> finish ()
  in

  (* Allocate and initialize a tuple with the fields es
     where dest is a rhs_finish.
     Continue with cont : (reg * ycon list * layout)-cont.
  *)
  let tuple_aux (env : env) es dest cont =
    let no_pathes env es = List.fold_left (fun no_pathes e ->
      let ((rawe,clist),_) = cogen_val env e in
      (not (is_path rawe)) & no_pathes) true es
    in
    let layout = get_layout env in
    let (rdest, spill1, layout1) = get_dest layout dest in
    let (rtemp,spill2,layout2) = 
      if no_pathes (set_layout env layout1) es then
	(Virt (new_ident ()), [], layout1)
      else 
	get_reg_not layout1 [rdest] 
    in
    (* Generate values *)
    let env' = set_layout env layout2 in
    let (es',cs') = List.fold_right (fun e (es',cs') -> 
      let (e',c') = cogen_val env' e in
      (e'::es', c'::cs')) es ([],[])
    in
    let rescs = cogen_cons cs' in
    let (il, cb) = cont (rdest, cs',layout2) in
    let (sz, initialization) = 
      List.fold_left2 (fun (sz,il) ((e,clist) as e') c ->
	let newsz = ysizeof delta c + sz in
	if is_path e then
	  (newsz, 
	   Mov (Reg rtemp, e') ::
	   Mov (Prjr (raw rdest, sz), raw (Reg rtemp)) :: il)
	else
	  (newsz, Mov (Prjr (raw rdest, sz), e') :: il)
	    ) (0, il) es' cs'
    in
    (* Allocate instruction *)
    let allocate = 
      Malloc (rdest, 
	      sz,
	      Mprod (List.map (fun c -> (c,Read)) rescs))
    in
    (spill1 @ (spill2 @ (allocate :: initialization)), cb)
  in
  (* Actual generation of code for simple right-hand sides. *)
  match rhs with
    Yprim (p,cs,es) -> cogen_prim (p,cs,es) env dest cont
  | Ytuple es ->
      (match es with
	[] -> cogen_prim (Yunit, [], []) env dest cont
      |	_ -> 
	  tuple_aux env es dest 
	    (fun (r,cs,layout) -> cont (YCprod cs, layout)))
  | Yproj (i,e) ->
      let ((raw_e, clist) as e',c') = cogen_val env e in
      (match c' with
	YCprod cs ->
	  let offset = get_offset delta cs i in
	  let (rdest, spill_i, source, layout') =
	    match raw_e with
	      Addr l ->
		let (r,spill_i,layout') = get_reg layout in
		(r,spill_i, raw(Prjl ((l,clist), offset)), layout')
	    | Reg r ->
	      	(match dest with
	    	  Bind_id x -> 
		    if is_dead deadvars x e then 
		      (r, [], raw (Prjr ((r,clist), offset)), layout)
		    else
		      let (r',spill_i,layout') = get_reg_not layout [r] in
		      (r', spill_i, raw (Prjr ((r,clist), offset)), layout')
	      	| Bind_loc r' ->
		    let (spill_i, layout') = spill_reg layout r' in
		    (r', [], raw (Prjr ((r,clist), offset)), layout'))
	    | Prjr ((r,_),_) ->
		let (r',spill_i, layout') = get_reg_not layout [r] in
		(r', spill_i @ [Mov (Reg r', e')], raw (Prjr (raw r', offset)),
 		 layout')
	    | Prjl (_,_) ->
		let (r',spill_i, layout') = get_reg layout in
		(r', spill_i @ [Mov (Reg r', e')], raw (Prjr (raw r', offset)),
 		 layout')
	    | _ -> die "cogen_simple_rhs: can't project off this"
	  in
	  let layout' =
	    match dest with
	      Bind_id x -> add_id layout' x (Rloc rdest)
	    | Bind_loc _ -> layout'
	  in
	  let (il, cbs) = cont (List.nth cs i, layout') in
	  (spill_i @ (Mov (Reg rdest, source) :: il), cbs)
      |	_ -> die "cogen_simple_rhs: projection of non-tuple type")

  (* NOTE:  Allowing pack of rhs *)
  | Ypack (rhs, v, k, c1, c2) ->
      (* Ypack (rhs, v, k, c1, c2)  == pack <rhs, c2> as E(v:k).c1 BUT:
	 Pack (e,c1,c2) == pack <e,c1> as c2
      *)
      let normc1 = ynorm kappa c1 in
      let normc2 = ynorm kappa c2 in
      let (r, layout_modifier) = 
	match dest with
	  Bind_id x ->
	    let r = 
	      (* this sucks but I don't know what to do in this case.
		 What I really want is for cogen_simple_rhs to pass
		 the register it uses as a destination to the continuation
		 and figure out which register to coerce from there 
	      *)
	      match get_free_reg layout with
	      	None -> Eax
	      |	Some r -> r
	    in 
	    (r, (fun layout -> add_id layout x (Rloc r)))
	| Bind_loc r -> (r, (fun x -> x))
      in
      let cont' = fun (_,layout) -> 
	let (il,cbs) = cont (YCexist (v,k,normc1), layout_modifier layout) in
	(Coerce (coerce (raw r) 
		   (Pack (cogen_con normc2, 
			  Cexist(v, cogen_kind k, cogen_con normc1)))) ::
	 il, cbs)
      in
      cogen_simple_rhs env rhs (Bind_loc r) cont'
  | Ytag (i, css, es) -> 
      if Yanormal.tag_is_value (i,css,es) then
	value_aux env rhs dest cont
      else
	(* the Tal tag for the ith element of the ycalculus sum *)
	let tag = get_tag css i in
	(* Translate constructors *)
	let normcss = List.map (List.map (ynorm kappa)) css in
	let normresc = YCsum normcss in
	let talresc = cogen_con normresc in
	(match (talresc, es) with
	  (_, []) ->
	    (* The sum is represented as a tag *)
	    die "tag should be value"
	| (Csum {sum_tags = _; sum_vt = Some (Tuple cs)}, _) ->
	    (* The sum is represented as a tuple *)
	    tuple_aux env es dest (fun (r,c,layout) -> 
	      singlei (Coerce (coerce (raw r) (Tosum talresc)))
		(normresc,layout) cont)

	| (Csum {sum_tags = st; sum_vt = Some (Variants vs)}, _) ->
	    (* The sum is represented as a variant (tag as 1st field of tuple) *)
	    let no_pathes env es = List.fold_left (fun no_pathes e ->
	      let ((rawe,clist),_) = cogen_val env e in
	      (not (is_path rawe)) & no_pathes) true es 
	    in
	    let layout = get_layout env in
	    let (rdest, spill1, layout1) = get_dest layout dest in
	    let (rtemp,spill2,layout2) = 
	      if no_pathes (set_layout env layout1) es then
	  	(Virt (new_ident ()), [], layout1)
	      else 
	  	get_reg_not layout1 [rdest] in
            (* Generate values *)
	    let env' = set_layout env layout2 in
	    let (es',cs') = List.fold_right (fun e (es',cs') -> 
	      let (e',c') = cogen_val env' e in
	      (e'::es', c'::cs')) es ([],[])
	    in
	    (* Generate continuation *)
	    let (il, cbs) = cont (normresc,layout2) in
            (* Initialize Instructions *)
	    let (sz, initialization) = 
      	      List.fold_left2 (fun (sz,il) ((e,clist) as e') c ->
	  	let newsz = ysizeof delta c + sz in
	  	if is_path e then
		  begin
		    (match rtemp with 
		      Virt _ -> die "cogen_simple_rhs: no_pathes error"
		    | _ -> ());
		    (newsz, 
		     Mov (Reg rtemp, e') ::
		     Mov (Prjr (raw rdest, sz), raw (Reg rtemp)) :: il)
		  end
	  	else
		  (newsz, Mov (Prjr (raw rdest, sz), e') :: il)
	     ) (wordsize, 
	   	Coerce (coerce (raw rdest) (Tosum talresc)) :: il) es' cs'
	    in
            (* Allocate Instruction: *)
	    let rescs = cogen_cons cs' in
	    let allocate = 
      	      Malloc (rdest, 
		      sz,
		      Mprod ((Cprim (PCtag tag), Read) :: 
			     List.map (fun c -> (c,Read)) rescs))
	    in
	    let tag_init = Mov (Prjr (raw rdest, 0), raw (Reg rtemp)) in
	    (spill1 @ (spill2 @ (allocate :: tag_init :: initialization)), 
	     cbs)
	| _ -> die "cogen_simple_rhs: tag_is_value error"
	)     
  | Ybuildexn (exnname, arg) ->
      let (rdest, spill1, layout') = get_dest layout dest in
      let env' = set_layout env layout' in
      let ((rawarg,clist1) as arg',argc) = cogen_val env' arg in
      let ((rawexnname,clist2) as exnname, exnnamec) = cogen_val env' arg in
      let size_arg = ysizeof delta argc in
      let size_exnname = ysizeof delta exnnamec in
      let malloc_instruction =
	Malloc (rdest, size_arg + size_exnname, Mprod
		[(cogen_con exnnamec, Read); (cogen_con argc, Read)])
      in
      if is_path rawarg or is_path rawexnname then
	let (rtemp, spill2, layout'') = get_reg layout' in
	let env' = set_layout env' layout'' in
	let (arg',_) = cogen_val env' arg in
	let (exnname',_) = cogen_val env' arg in
      	let (il, cbs) = cont (YCconst YCexn, layout'') in
	let instructions =
	spill1 @
      	(malloc_instruction ::
	(spill2 @
	((if is_path rawexnname then
	  Mov (Reg rtemp, exnname') ::
      	  Mov (Prjr (raw rdest, 0), raw (Reg rtemp)) :: []
	else 
	  [Mov (Reg rdest, exnname')]) @
	((if is_path rawarg then
	  Mov (Reg rtemp, arg') ::
	  Mov (Prjr (raw rdest, size_exnname), raw (Reg rtemp)) :: []
	else
      	  [Mov (Prjr (raw rdest, size_exnname), arg')]) @
      	  Coerce (coerce (raw rdest) Toexn) :: il))))
	in
	(instructions, cbs)
      else
	let (il, cbs) = cont (YCconst YCexn, layout') in
	(spill1 @
	(malloc_instruction ::
	 Mov (Prjr (raw rdest, 0), exnname) ::
	 Mov (Prjr (raw rdest, size_exnname), arg') ::
	 Coerce (coerce (raw rdest) Toexn) :: il), cbs)
  | Ycons (e1,e2) ->
      tuple_aux env [e1;e2] dest (fun (r,cs,layout) ->
 	(match cs with
	  [c1; c2] ->
	    singlei (Coerce (coerce (raw r) (Roll (cogen_con c2)))) 
	      (c2,layout) cont
      	| _ -> die "cogen_simple_rhs: Ycons")
	  )

  (* the only other valid right-hand-sides are values *)
  | _ -> value_aux env rhs dest cont

(* Code Generation for right-hand sides of Anormal form.
   env: explained above.
   next: type finish, see below

   results:
     a list of instructions which perform the rhs computation.
     a list of code blocks which must be added to the heap.
*)

type finish =
    Return              (* last instruction should return result in Eax *)
  | Bind of ident       (* result of computation is bound to x *)
  (* if the computation ends with a jump or a fall through, it
     should be to the following label with the result in register given
   *)
  | Jump of reg * label * con list * (regs_type * frame_type)
  (* result of computation should be put in loc
     and the computation should Jmp to label,
     applying the con list left to right.
     Computation must end in state given by regs_type, frametype
  *)
  | Fall of reg * label * con list * (regs_type * frame_type)
  (* result of computation should be put in loc
     and computation may Fallthru to label,
     applying the con list left to right. 
     Computation must end in state given by regs_type, frametype
  *)

let rec cogen_rhs 
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env : env) 
    next rhs cont
    : (instruction list * code_blockl list) =

  (* push arguments onto stack and save registers.
     return instruction list and new environment
  *)
  let push_all
      (((delta, gamma, kappa) as ctxt,
      	((ra, regs, frame, ret, deadvars) as layout),
      	((rho1, rho2) as taldelta)) as env : env)
      args argsizes dyingvars =
    (* argument onto stack, spilling where necessary *)
    let push_arg (env, offset, il) arg thisargsz =
      let ((g,clist) as tal_arg,_) = cogen_val env arg in
      let layout = get_layout env in
      let dest = Prjr (raw Esp, offset) in
      let (il, layout') = 
      	if valid_binops g dest then
      	  (Mov (dest, tal_arg) :: il, layout)
      	else
	  let (r, spill_i, layout') = get_reg layout in
	  (spill_i @
	   Mov (Reg r, tal_arg) ::
	   Mov (dest, raw (Reg r)) ::
	   il, layout')
      in
      (set_layout env layout', offset + thisargsz, il)
    in
    let push_regs layout =
      let rec loop i (il, layout) = 
      	if i < 6 then
	  let slot = match layout with (_,regs,_,_,_) -> regs.(i) in
	  match slot with
	    None -> loop (i+1) (il, layout)
	  | Some x ->
	      if Set.member dyingvars x then
		loop (i+1) (il, layout)
	      else
      	  	let (spill_i, layout) = spill_reg layout (int2reg i) in
	  	loop (i+1) (spill_i @ il, layout)
      	else
	(il, layout)
      in
      loop 0 ([], layout)
    in
    let (env, _, push_args_il) = 
      List.fold_left2 push_arg (env, 0, []) args argsizes 
    in
    let (push_regs_il, layout') = push_regs (get_layout env) in
    (push_args_il @ (List.rev push_regs_il), set_layout env layout)
  in
  (* instantiate a function *)
  let make_instantiation 
    ((((delta, gamma, kappa) as ctxt),
    ((ra, regs, frame, ret, deadvars) as layout),
    ((rho1, rho2) as taldelta)) as env)
    normtyargs lab =
    let taltyargs = cogen_cons normtyargs in
    let frame_type = Cappend (frame2type env, Cvar rho1) in
    (* Apply type arguments. *)
    let instantiated_label =  
      List.fold_left (fun l c -> coerce l (Tapp (cogen_con c))) 
	lab normtyargs
    in
    (* Apply stack type arguments.  The exception handler remains unchanged. *)
    coerce (coerce instantiated_label (Tapp (frame_type))) (Tapp (Cvar rho2))
  in
  (* Code is called as follows:
     1. Push all arguments right to left
     2. Push all registers in context onto the stack
     3. Call function
     4. Save Result
  *)
  let app_aux  
      (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env : env) 
      ((f,tyargs,args) : yterm * ycon list * yterm list)
      (dyingvars : ident Set.set)
      (rdest : reg) 
      cont = 

    (* Type Checking *)
    (* get the type of f first *)
    let (_, fc) = cogen_val env f in
    (* normalized result type, normalized type arguments *)
    let (normresc, normtyargs) =
      match fc with
	YCabs (vks, cs, c) ->
	  let (kappa', normtyargs) =
	    List.fold_right2 (fun (v,k) tyarg (kappa, normtyargs) ->
	      let normtyarg = ynorm kappa tyarg in
	      (Dict.insert kappa v normtyarg, normtyarg :: normtyargs) 
	    ) vks tyargs (kappa, []) 
	  in
	  (ynorm kappa' c, normtyargs)
      |	_ -> die "cogen_rhs: application of non-function"
    in
    (* argsz: total size of all arguments, argsize: individual sizes of args *)
    let argsz, argsizes = 
      match fc with
      	YCabs(vks, cs, c) ->
	  let delta = 
	    List.fold_left (fun delta (v,k) -> Dict.insert delta v k) 
	      delta vks 
	  in
	  List.fold_right (fun c (argsz, argsizes) ->
	    let sz = ysizeof delta c in
	    (sz + argsz, sz::argsizes)) cs (0,[])
      |	_ -> die "non-function in app"
    in

    (* Code Generation *)
    (* Allocate and initialize new frame with contents of old frame *)
    let oldframesz = Array.length frame in
    let newframesz = (argsz/4 + oldframesz) in 
    let frame' = new_frame newframesz in
    Array.blit frame 0 frame' 0 oldframesz;
    
    (* Spill registers and place arguments on stack *)
    let env' = set_layout env (ra, regs, frame', ret, deadvars) in
    let (push_instructions, env') = push_all env' args argsizes dyingvars in 
    let frame' =
      match get_layout env' with
      	(_,_,frame',_,_) -> frame'
    in 
    Array.blit frame' 0 frame 0 oldframesz;
    (* Generate call *)
    (* Construct the coercion to apply to the label *)       
    let (f', _ ) = cogen_val env' f in
    let instantiated_label = 
      make_instantiation env normtyargs f'
    in
    let (il, cbs) = cont (normresc, get_layout env) in
    let call = 
      if rdest = Eax then
	Call instantiated_label :: il
      else
      	Call instantiated_label ::
      	Mov (Reg rdest, raw (Reg Eax)) :: il
    in
    ((ArithBin(Sub,Reg Esp,Immed argsz))::
     (push_instructions @ call), cbs)
  in
  (* Tail Calls:
     Code is called as follows:
     1. Copy stack arguments into registers if they are not in the right place.
     2. Stack alloc or stack free so the stack is sized properly for arguments.
     3. Coerce stack to appropriate type.
     4. Mov all arguments onto stack.
     5. Push return address.
     5. Call function
  *)

    (* The rearrangement of arguments on the stack and the return address to their
       correct positions.  Yup.  It's pretty hectic.
    *)
    let relocate args rtemp1 rtemp2 argsz 
 	((ra, regs, oldframe, ret, deadvars) as layout) =
      (* Make sure there is enough space to move the arguments around.
	 Allocate more space if the arguments take up more space. *)
      let newframesz = (argsz + wordsize)/4 in
      let oldframesz = Array.length oldframe in 
      let spill0, layout = 
	if newframesz > oldframesz then
	  begin
	    let newframe = new_frame (newframesz) in
	    Array.blit oldframe 0 newframe 0 oldframesz;
	    ([ArithBin(Sub,Reg Esp,Immed ((newframesz-oldframesz)*4))], 
	     (ra, regs, newframe, ret, deadvars))
	  end
	else
	  ([], layout)
      in 
      let arg_array = Array.of_list (List.map (fun a -> Some a) args) in	 
      (* True if the return address has been relocated *)
      let moved_return = ref false in
      (* Find the slot corresponding to y *)
      let get_slot y = 
	let rec get_id arg =
	  match arg with
	    Yvar x -> Some x
	  | Ypack (a, _,_,_,_) -> get_id a
	  | Yfold (_,_,_, a) -> get_id a
	  | Yunfold (_,_,_,a) -> get_id a
	  | Yroll (_,a) -> get_id a
	  | Yunroll a -> get_id a
	  | _ -> None
	in
	let i = ref 0 in
	let notfound = ref true in
	let slot = ref (-1) in
	while (!notfound) && (!i < (Array.length arg_array) - 1) do
	  match arg_array.(!i) with
	    None -> incr i
	  | Some arg ->
	      if get_id arg = Some y then (slot := !i; notfound := false)
	      else incr i
	done;
	if !notfound then None
	else Some (!slot)
      in
      let get_offset framesz slot =
	match sloc2genop framesz (4*slot) with
	  Prjr (_,j) -> j
	| _ -> die "get_slot: impossible"
      in
      (* Recursively move arguments to their correct positions on the
	 stack.  Filled_reg contains the argument to be moved.  It must
	 be moved to the stack slot "toslot". Empty_reg
	 is a free register.
      *)
      let rec bubble filled_reg empty_reg toslot 
	  (spill_cont, ((ra, regs, frame, ret, deadvars) as layout)) =
	let newoffset = get_offset (Array.length frame) toslot in
	let dest = raw (Prjr ((Esp, []), newoffset)) in
	match frame.(toslot) with
	  Filled_slot y ->
	    let yslot = get_slot y in
	    (match yslot with
	      Some yslot ->
	    	arg_array.(yslot) <- None;
	    	bubble empty_reg filled_reg yslot
		  ((fun spill_i ->
		    spill_cont
		      (Mov (Reg empty_reg, dest) ::
		       Mov (fst dest, raw (Reg filled_reg))::spill_i)),
		    layout)
	    | None ->
		(* If it is not an argument, we can overwrite it *)
		frame.(toslot) <- Empty_slot (Cprim (PCtag (-1)));
		bubble filled_reg empty_reg toslot
		  (spill_cont, layout))
	| Empty_slot c ->
	    ((fun spill_i -> 
	      spill_cont
		(Mov (fst dest, raw (Reg filled_reg))::spill_i)), layout)
	| Anon_slot c ->
	   (* Contains return address *)
	    moved_return := true;
	    bubble empty_reg filled_reg (newframesz - 1)
	      ((fun spill_i ->
		spill_cont
		  (Mov (Reg empty_reg, dest) ::
		   Mov (fst dest, raw (Reg filled_reg))::spill_i)),
		layout)
      in
      (* Relocate arguments *)
      let reloc i arg (spill_cont, ((ra, regs, frame, ret, deadvars) as layout)) =
	let ((rawarg, clist) as talarg,_) = 
	  cogen_val (set_layout env layout) arg in
        match rawarg with
	  Prjr ((Esp,_), k) ->
	    (* We are going to move this object. *)
	    arg_array.(i) <- None;
	    (* Find the new slot number and offset (extra -1 for ret addr) *)
	    let newslot = newframesz - i - 2 in
	    let newoffset = get_offset (Array.length frame) newslot in
	    (* Find the object name *)
	    let oldslot = (Array.length frame) - 1 - (k/4) in
	    let x = match frame.(oldslot) with 
	      Filled_slot x -> x
	    | _ ->
		die "tail_call: must be moving id"
	    in
	    if oldslot = newslot then
	      (* Argument already in the correct spot *)
	      (spill_cont, layout) 
	    else
	      begin
	      (* object destination on stack *)
	      let dest = (Prjr ((Esp,[]), newoffset), []) in
	      let (_, regs, frame, _, _) = layout in
	      match frame.(newslot) with
		Filled_slot y ->
		  if y = x then (spill_cont, layout) 
                    (* Object already in correct spot *)
		  else 
                    (* Another object in designated slot. Move it out. *)
		    begin
		      regs.(reg2int rtemp2) <- Some y;
		      frame.(oldslot) <- 
			 Empty_slot (cogen_con (Dict.lookup gamma x));
		      let yslot = get_slot y in
		      (match yslot with
			None -> 
			  (* if it is not an argument, it can be overwritten *)
			  frame.(oldslot) <- 
			     Empty_slot (cogen_con (Dict.lookup gamma x));
			  ((fun spill_i ->
			    spill_cont
			      (Mov (Reg rtemp1, talarg) ::
			       Mov (fst dest, raw (Reg rtemp1)) :: 
			       spill_i)), layout)
		      |	Some yslot ->
			  bubble rtemp2 rtemp1 yslot
			    ((fun spill_i ->
		      	      spill_cont
			  	(Mov (Reg rtemp1, talarg) ::
				 Mov (Reg rtemp2, dest) ::
				 Mov (fst dest, raw (Reg rtemp1)) ::
				 spill_i)), layout))
		    end
	      |	Empty_slot c ->
		  frame.(oldslot) <- Empty_slot (cogen_con (Dict.lookup gamma x));
		  ((fun spill_i ->
		    spill_cont
		      (Mov (Reg rtemp1, talarg) ::
		       Mov (fst dest, raw (Reg rtemp1)) :: spill_i)), layout)
	      |	Anon_slot c ->
		  (* Return address is in this slot. Move it to top of stack. *)
		  begin
		      frame.(oldslot) <- 
			 Empty_slot (cogen_con (Dict.lookup gamma x));
		      bubble rtemp2 rtemp1 ((Array.length frame) - 1)
		      ((fun spill_i ->
		      	spill_cont
			  (Mov (Reg rtemp1, talarg) ::
			   Mov (Reg rtemp2, dest) ::
			   Mov (fst dest, raw (Reg rtemp1)) ::
			   spill_i)), layout)
		  end
	      end
	| _ -> (spill_cont, layout)
      in
      let (spill_cont, layout) = 
      	vector_foldall (fun i arg_opt (spill_cont,layout) ->
	  match arg_opt with
	    None -> (spill_cont, layout)
	  | Some arg -> reloc i arg (spill_cont, layout)) arg_array 
	((fun spill_i -> spill_i), layout)
      in
      if !moved_return or ret = newframesz - 1 then
      	(layout, spill0 @ (spill_cont []), arg_array)
      else
	let frame = match layout with (_,_,frame,_,_) -> frame in
      	let framelength = (Array.length frame) in
	let source = raw (sloc2genop framelength (ret*4)) in 
	let dest = Prjr ((Esp,[]), 4*framelength - 4*newframesz) in
(*	let dest = sloc2genop framelength (4*newframesz) in *)
	(layout, 
	 spill0 @ (spill_cont 
	    (Mov (Reg rtemp1, source) :: 
	     Mov (dest, raw (Reg rtemp1)) :: [])), 
	 arg_array)
    in
    
  let tail_call_aux
      (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env : env)
      ((f,tyargs,args) : yterm * ycon list * yterm list) 
      (cont : ycon * layout -> (instruction list * code_blockl list)) =
    (* Type Inference *)
    (* Function Type: *) 
    let (f', fc) = cogen_val env f in
    (* Determine normalized ycalculus result type *)
    let (normresc, normtyargs) =
      match fc with
	YCabs (vks, cs, c) ->
	  let (kappa', normtyargs) =
	    List.fold_right2 (fun (v,k) tyarg (kappa, normtyargs) ->
	      let normtyarg = ynorm kappa tyarg in
	      (Dict.insert kappa v normtyarg, normtyarg :: normtyargs) 
	    ) vks tyargs (kappa, []) 
	  in
	  (ynorm kappa' c, normtyargs)
      |	_ -> die "cogen_rhs: application of non-function"
    in
    (* argsz: total size of all arguments, argsize: individual sizes of args *)
    let argsz, argsizes = 
      match fc with
      	YCabs(vks, cs, c) ->
	  let delta = 
	    List.fold_left (fun delta (v,k) -> Dict.insert delta v k) 
	      delta vks 
	  in
	  List.fold_right (fun c (argsz, argsizes) ->
	    let sz = ysizeof delta c in
	    (sz + argsz, sz::argsizes)) cs (0,[])
      |	_ -> die "non-function in app"
    in
    (* Move f into register if it is on the stack *)
    let (f', rlist, spill0, layout) =
      match f' with
      	(Prjr ((Esp, _), i), _) -> 
	  let (r, spill_i, layout) = get_reg layout in
	  (raw (Reg r), [r], spill_i @ [Mov (Reg r, f')], layout)
      | _ -> (f', [], [], layout)
    in
    (* Apply stack type arguments. *)
    let frame_type = Cvar rho1 in
    let instantiated_label = 
      coerce (coerce f' (Tapp frame_type)) (Tapp (Cvar rho2)) in
    (* Apply type arguments. *)
    let instantiated_label =
      List.fold_left (fun l c -> 
	coerce l (Tapp (cogen_con c))) instantiated_label normtyargs
    in

    (* Relocate all arguments on stack and return address *)
    (* Get 2 free registers to use as temporaries *)
    let (rtemp1,spill1,layout) = get_reg_not layout rlist in
    let (rtemp2,spill2,layout) = get_reg_not layout rlist in
    let oldframesz = match layout with (_,_,frame,_,_) -> 4*(Array.length frame) in
    let (layout, relocate_instructions, arg_array) = 
      relocate args rtemp1 rtemp2 argsz layout
    in   
    (* Here follows the call sequence instructions built from last to first *)
    (* Continuation - instructions are dead code *)
    let (_, cb) = cont (normresc, layout) in
    (* Instructions to call function *)
    let instructions = Jmp instantiated_label :: [] in
    (* Instruction to smalloc or sfree and coerce stack *)
    let instructions =
      let diff = oldframesz - (argsz + wordsize) in
      if diff <= 0 then
	(* Any extra space already added by relocation *)
	instructions
      else
	(* Arguments need less space than currently allocated *)
      	ArithBin (Add, Reg Esp, Immed diff) :: 
        instructions
    in  	
    (* Instructions to push arguments on stack *)
    let env = set_layout env layout in
    let diff = match layout with (_,_,frame,_,_) -> 
      4*(Array.length frame) - (argsz + wordsize)
    in
    let instructions =
      vector_foldall (fun i arg_opt il -> 
	match arg_opt with
	  None -> 
            (* Yippee.  The arg is already in the right place *)
	    il
	| Some arg ->
	    (* if the argument is a path, we need to do 2 moves to get it
	       onto the stack.
 	    *)

	    let (rawarg, clist) as talarg, _ = cogen_val env arg in  
	    let offset = diff + (i+1) * 4 in
	    if is_path rawarg then
	       Mov (Reg rtemp1, talarg) ::
	       Mov (Prjr (raw Esp, offset), raw (Reg rtemp1)) :: il
	    else	  
	       Mov (Prjr (raw Esp, offset), talarg) :: il
       ) arg_array instructions
    in
    (spill0 @ spill1 @ spill2 @ relocate_instructions @ instructions, cb)
  in

  (************** transformation of cases: *****************)
  (*
     case v of 
       i1 -> x1.e1
     | i2 -> x2.e2
     ...
     | in -> xn.en
     | default -> ed

     is transformed into:
     
       Mov rtemp, v
       Cmp rtemp i1
       Jeq L1 
       Cmp rtemp i2
       Jeq L2
       ...
       Cmp rtemp i(n-1)
       Jeq L(n-1)
       Code for default;
       Jump as given by jbranch
   Ln: Code for en;
       Jump as given by jbranch
   L2: Code for e2;  
       Jump as given by jbranch
   L1: Code for e1;
       Fallthru (or jump or return as given by fbranch)
next:

     ...
     
     If we are casing on a sum, we handle it similarly.
   *)

   (* Separate the case branches which switch off a sum tag from the cases
      which must project the tag from the tuple. 
      
      Parameters:
      normcss: normalized ycon list list giving the type of the sum
      cases: list of case branches

      Results:
      are_variants: boolean: true <=> there are 2 or more sum elements which
                                      carry values
                             false otherwise
      (const_cases, carry_cs, carry_cases):
        const_cases: case branches which case on a Tag.
        carry_cs: types of arguments which case on a carried value.
        carry_cases: case branches which case on a carried value.
      - Each case is rewritten so that it switches on the appropriate
        tag.  ie: if our sum is: () + (int) + (int) + () then 
                     the case (1,x,e) -> (0,x,e)
                  reflecting the fact that an element in the second component
                  of the sum will be represented using a variant with a 0 tag
           Similarly, case (3,x,e) -> (1,x,e) as the last element of the
                  sum will be represented as Tag(1)
   *)
  let sort_cases normcss cases =
    let are_variants = ref false in
    (* Create a mapping from ycalculus sum indices to tal indices and 
       whether or not that particular element of the sum is a simple 
       tag (true) or is a tuple/variant(false)
    *)
    let rec mk_tagDict normcss tagnum vtnum i =
      match normcss with
	[] -> (if vtnum > 1 then are_variants := true; Dict.empty compare)
      | []::normcss -> 
	  Dict.insert 
	    (mk_tagDict normcss (tagnum+1) vtnum (i+1)) i (tagnum,true) 
      | hd::normcss -> 
	  Dict.insert 
	    (mk_tagDict normcss tagnum (vtnum+1) (i+1)) i (vtnum,false)
    in
    (* Divide cases into cases on a simple tag(const_cases) and cases on
       a tuple/variant (carry_cases). Each cases given the correct Tal 
       tag as indicated by tagDict.
    *)
    let sort_cases tagDict cases =
      let aux (i,x,e) (const_cases, carry_cs, carry_cases) =
	  let (i', is_tag) = Dict.lookup tagDict i in 
	  if is_tag then 
	      ((i',x,e)::const_cases, carry_cs, carry_cases)
	  else
	    if !are_variants then
	      	(const_cases, 
		 YCprod (YCsingle i' :: List.nth normcss i) :: carry_cs,
		 (i',x,e)::carry_cases)
	    else
	      	(const_cases, YCprod (List.nth normcss i) :: carry_cs, 
		 (i',x,e)::carry_cases)
      in
      List.fold_right aux cases ([],[],[])
    in
    let cases = Sort.list (fun (i,_,_) (j,_,_) -> i <= j) cases in
    let tagDict = mk_tagDict normcss 0 0 0 in
    (!are_variants, sort_cases tagDict cases)
  in

  (* Generate codetype (sequence of foralls followed by code) 
     from everything in the context *)
  let cogen_codetype 
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env) =
    Cforall (rho1, Kstack, 
      Cforall (rho2, Kstack, 
	Dict.fold_dict (fun v k c -> 
	  Cforall (v, cogen_kind k, c)) delta (Ccode (regs2reg_state env))))
  in
  (* split_codetype:
       Factor normalized code type (a sequence of foralls followed by code)
       into the (variable,kind) list and register_state
     make_codetype: the inverse
  *)
  let split_codetype c =
    let rec sc c =
      match c with
	Ccode gamma -> ([],gamma)
      |	Cforall (v,k,c) -> let (vks,gamma) = sc c in ((v,k)::vks, gamma)
      |	_ -> die "split_code:  not code type"
    in
    sc c
  in
  let make_codetype vks gamma = 
    List.fold_right (fun (v,k) c -> Cforall (v,k,c)) vks (Ccode gamma)
  in
  (* Partially apply tal types cs to argument *)
  let apply_types (arg,clist) cs =
    (arg, List.fold_left (fun clist c -> Tapp c :: clist) clist cs)
  in
  (* Given a label coercion l, extract type variables from environment 
     (including stack variables) and apply to l 
  *)
  let coerce_l delta rho1 rho2 l =
    apply_types l 
      (rho1::rho2::(Dict.fold_dict (fun v _ l -> Cvar v :: l) delta []))
  in

  (* Given a case statement, generate code for the case statement.
     rcase: the register to switch on - it can be trashed
     rc: its normalized ycalculus type.  
     cases: the branches of the case statement 
     opt: the optional default.
     jbranch (jump branch) and fbranch (fall_thru branch) are of type finish.
  *)
  let case_aux 
      (((delta, gamma, kappa) as ctxt,
      	((ra, regs, frame, ret, deadvars) as layout),
      	((rho1, rho2) as taldelta)) as env : env) 
      (rcase, rc, cases, opt) jbranch fbranch cont =
    (* kill variables dead immediately upon entry to the branch of the
       case statement with bound variable x *)
    let kill_at_entry env cases x loc =
      (* kill variables after switch *)
      let dv = 
	match cases with 
	  [] -> Set.empty compare_idents
	| (_,x,_) :: cases -> get_deadvars_case deadvars x 
      in
      let layout = kill_vars env dv in
      (* kill variables on entry to the particular branch.  Add x at location
	 loc if it is not dead. *)
      let dv = get_deadvars deadvars x in
      let (dv, layout) = 
	if Set.member dv x then (Set.delete dv x, layout)
	else (dv, add_id layout x loc)
      in
      kill_vars (set_layout env layout) dv 
    in
     (* do_case: Generate code for one of the branches of a case statement.
 
	Parameters:
	Arg 1: (i,x,e): a branch of a case statement
	Arg 2: c: the type of the variable which is bound
        Arg 3: a tuple:
	  (rcase, is_sum):
	    rcase: the register on which we perform the case
	    is_sum: true if casing on a sum
  	  next: how to finish this branch of the case.
	  ls: a list of coerced labels for the other branches of the case.
	  cse_cont: a continuation containing following instructions

	Implicitly uses: label_state, env

	Return:
	(rcase, is_sum): unchanged, as above
	next': how to finish.  A jump is always returned.
	ls': the coerced label for this code block followed by ls
	cse_cont': the continuation performing the instructions of this
	   case branch and following with cse_cont 
    *)
    let do_case 
	((i,x,e) : int * ident * yterm)
 	((c,tc) : ycon * con)
	((rcase, is_sum), next, ls, cse_cont) =
      let env = set_layout env
	  (ra, copy_regs regs, copy_frame frame, ret, deadvars) in
      (* add x to context: *)
      let gamma' = Dict.insert gamma x c in
      (* eliminate dead variables from context *)
      let layout = kill_at_entry (set_layout env layout) cases x (Rloc rcase) in
      let codetype = cogen_codetype (set_layout env layout) in
      let ctxt = (delta, gamma', kappa) in
      let (il', cbs') = cogen_exp (ctxt, layout, taldelta) next e cse_cont in
      let l = new_label "cse" in
      (* The first int case can fall through but the rest must jump to l *)
      let next =
	match next with
	  Fall (x,l,cs,memory) -> Jump (x,l,cs,memory)
	| _ -> next
      in
      ((rcase, is_sum),
       next, 
       coerce_l delta (Cvar rho1) (Cvar rho2) (raw l) :: ls,
       (* case continuation *)
       fun _ ->
	 if is_sum & not (is_dead deadvars x (Yvar x)) then
	   ([], (l, codetype, 
		 Coerce (coerce (raw rcase) Fromsum)::il') :: cbs')
	 else
	   ([], (l, codetype, il') :: cbs'))
    in
    (match rc with
      YCconst YCint ->
	(* Generate continuation which produces code for each branch: *)
	let (_,_,ls, cases_cont) = 
	  List.fold_right2 do_case cases 
	    (List.map (fun _ -> (YCconst YCint, byte4)) cases) 
	    ((rcase, false), fbranch, [], cont)
	in
	(* Generate code for default: *)
 	let (defil, defcbs) =
	  match opt with
	    None -> 
	      die "integer case with no default"
	  | Some e -> 
	      let (ra, regs, frame, ret, deadvars) = get_layout env in
	      let layout = 
		(ra, copy_regs regs, copy_frame frame, ret, deadvars) in
	      let env = set_layout env layout in
	      cogen_exp env jbranch e cases_cont
	in

	(* Generate comparison code *)
	let instructions = List.fold_right2 (fun (i,x,e) l il ->
	  Cmp (Reg rcase, Immed i) :: 
	  Jcc (Eq, l) :: il) cases ls defil
	in
	(instructions, defcbs)
    | YCsum css ->
	let normcss = List.map (List.map (ynorm kappa)) css in
	let (are_variants, (const_cases, carry_cs, carry_cases)) = 
	  sort_cases normcss cases in
	(* Get the default:
	    - if there is a default, use that
	    - otherwise use a carry_case
	    - otherwise use a const_case
        *)
	let rec last xs = 
	  match xs with
	    [] -> die "last: must be at least 1 element in list"
	  | [x] -> (x,[])
	  | x::xs -> let x',xs' = last xs in (x', x::xs')
	in
	let regs' = copy_regs regs in
	let frame' = copy_frame frame in
	let (const_cases, carry_cs, carry_cases, default, fst_instr, def_env) = 
	  match (const_cases, carry_cs, carry_cases, opt) with
	    ([],_,[],None) -> die "case_aux: empty sum case"
	  | (const_cases,_,[],None) ->
	      let ((i,x,e),const_cases) = last const_cases in
	      (const_cases, carry_cs, carry_cases, e,
	       (if is_dead deadvars x (Yvar x) then None
	       else Some (Coerce (coerce (raw rcase) Fromsum))),
	       let (delta,gamma,kappa) as ctxt =
	        (delta, Dict.insert gamma x (YCsingle i), kappa)
	       in
	       let layout = (ra, regs', frame', ret, deadvars) in
	       let layout = kill_at_entry (set_layout env layout) cases 
		   x (Rloc rcase) 
	       in
	       (ctxt,layout,taldelta))
	  | (_,carry_cs,carry_cases,None) ->
	      let ((i,x,e),carry_cases) = last carry_cases in
	      let (c, carry_cs) = last carry_cs in
	      (const_cases, carry_cs, carry_cases, e,
	       (if is_dead deadvars x (Yvar x) then None
	       else Some (Coerce (coerce (raw rcase) Fromsum))),
	       let (delta, gamma, kappa) as ctxt = 
		 (delta, Dict.insert gamma x c, kappa)
	       in
	       let layout = (ra, regs', frame', ret, deadvars) in
	       let layout = kill_at_entry (set_layout env layout) cases 
		   x (Rloc rcase)
	       in
	       (ctxt, layout, taldelta))
	  | (_,_,_,Some e) ->
	      (const_cases, carry_cs, carry_cases, e, None, 
	       (ctxt, (ra, regs, frame, ret, deadvars), taldelta))
	in	    
	(* Pair ycalculus and tal types bound to x in each tag case *)
	let cs = List.fold_right 
	    (fun (i,x,e) cs -> (YCsingle i, 
				(Csum {sum_tags = [i];sum_vt = None})) :: cs)
	    const_cases []
	in 
	(* Generate continuation which produces code for each branch: *)
	let (_, _, const_ls, cases_cont) = 
	  List.fold_right2 do_case
	    const_cases cs ((rcase, true), fbranch, [], cont) 
	in
	(* Pair ycalculus and tal types bound to x in each carry case *)
	let cs = List.fold_right2
	    (fun (i,x,e) yc cs -> 
	      let tc = cogen_con yc in
	      let tc = match tc with
		Cprod ((c,cap,init) :: cs) ->
		  Csum {sum_tags = []; sum_vt =
		       if are_variants then
			 Some (Variants [(i, cs)])
		       else
			 Some (Tuple ((c,cap,init) :: cs))}
	      |	Cprod [] -> die "carry case should not contain unit"
	      |	_ -> die "carry case translation not product"
	      in
	      (yc,tc)::cs) carry_cases carry_cs []
	in
	(* Generate code for carry cases *)
	let (_, _, carry_ls, cases_cont) =
	  List.fold_right2 do_case
	    carry_cases cs ((rcase, true), fbranch, [], cases_cont) 
	in
	let (defil, defcbs) =
	  cogen_exp def_env jbranch default 
	    cases_cont
	in
	let defil = 
	  match fst_instr with
	    None -> defil
	  | Some i -> i::defil
	in 
	(* Generate comparison code for tuples/variants *)
	let instructions =
	  if are_variants then
	    List.fold_right2 (fun (i,x,e) l il ->
	      Btagvar (rcase, i, l, BelowEq) :: il) 
	      carry_cases carry_ls defil
	  else
	    (* Case for a tuple.  We test using Btagi Above. *)
	    List.fold_right2 (fun (i,x,e) l il ->
	      Btagi (rcase, 255, l, Above) :: il) 
	      carry_cases carry_ls defil
	in
	(* Generate comparison code for constant constructors *)
	let instructions = List.fold_right2 (fun (i,x,e) l il->
	  Btagi (rcase, i, l, BelowEq) :: il) 
	    const_cases const_ls instructions
	in
	(instructions, defcbs)
    | _ -> die "cogen_rhs: case_aux: case on non-sum or int"
    )
  in
  (* Add coercions so that ctxt matches the jump target *)
  let jumpfall_coerce ((ra, regs, frame, ret, deadvars) as layout) 
      (regs',frame') icont =
    let (stack_coercions,il) = vector_foldall (fun i slot (clist,il) ->
      match slot, frame.(i) with
	(Empty_slot (Cprim PCjunk), Empty_slot (Cprim PCjunk)) -> (clist,il)
      |	(Empty_slot (Cprim PCjunk), _) -> (Slot (4*i) :: clist, il)
      |	(Filled_slot x, Filled_slot y) ->
	  if (compare_idents x y) = 0 then
	    (clist, il)
	  else
	    let loc = get_loc layout y in
	    begin
	      match loc with
	      	Rloc r -> 
		  let source = raw (Reg r) in
		  let dest = sloc2genop (Array.length frame) (4*i) in
	    	  (clist, Mov (dest,source) :: il)
	      |	_ ->
		  die "jumpfall_coerce: variable in unexpected location"
	    end
      |	_ -> (clist, il)
      ) frame' ([],icont)
    in
    Coerce (Esp, stack_coercions) :: il
  in
  (* Continuation for returning directly *)
  let return_cont =
     fun (c,((ra, regs, frame, ret, deadvars) as layout)) -> 
       let (il,cbs) = cont (c,layout) in
       (* Subtract arguments but not return address(done by default)
	  from stack 
       *)
       let decr_stack = 
	 match layout with
	   (_,_,frame,_,_) -> ret * 4
       in
       if decr_stack < 0 then
	 die "cogen_simple: popping negative bytes off stack during return"
       else
       (* before returning, must pop stack up to the return address *)
	 let frame_subtract = 4 * ((Array.length frame) - 1 - ret) in
	 if frame_subtract > 0 then
	   (ArithBin (Add, Reg Esp, Immed frame_subtract) :: 
	    [Retn (Some decr_stack)], cbs)
	 else
	   ([Retn (Some decr_stack)], cbs)
  in
  (* Continuation for fall thrus *)
  let fallthru_cont memory cs =
    fun (c,layout) ->
      let il = jumpfall_coerce layout memory [Fallthru cs] in
      let (_,cbs) = cont (c, layout) in
      (il, cbs)
  in
  (* Continuation for jumps *)
  let jump_cont memory l cs =
    let cl = apply_types (raw (Addr l)) cs in
    fun (c,layout) ->
      let il = jumpfall_coerce layout memory [Jmp cl] in
      let (_,cbs) = cont (c, layout) in
      (il,cbs)    
  in
  (* the continuation should remove all dead variables from the context *)
  let cont = 
    match next with
      Bind x -> 
	fun (c, layout) ->
	  let gamma = Dict.insert gamma x c in
	  let layout' = 
	    kill_vars ((delta,gamma,kappa), layout, taldelta) 
	      (get_deadvars deadvars x) 
	  in
	  cont (c, layout')
    | _ -> cont
  in
  match rhs with
    Yapp (f, cs, args) ->
      (match next with
    	Return -> 
	  (* This is a tail call. 
	     We don't need to save any registers and we can just 
	     put the return address on the stack and jump to the
	     the next function.  For now, we optimize fully.
	     We just won't save the registers. Notice that ssz
	     contains the size of the return address.  Since Retn
	     automatically pops off the return address, we use the
	     optional (ssz-4) for the arguments.

	  app_aux env (f,cs,args) (Set.empty compare_idents) Eax return_cont
	     *)
	  tail_call_aux env (f,cs,args) cont
      | Bind x  -> 
	  app_aux env (f,cs,args) (get_deadvars deadvars x) Eax
	    (fun (c,((_,_,_,_,deadvars) as layout)) -> 
	      let layout' = 
		if is_dead deadvars x (Yvar x) then layout
		else add_id layout x (Rloc Eax) 
	      in
	      cont (c, layout'))
      |	Jump (r,l,lcs,memory) ->
	  app_aux env (f,cs,args) (Set.empty compare_idents) r 
	    (jump_cont memory l lcs)

      |	Fall (r,l,lcs, memory) ->
	  app_aux env (f, cs, args) (Set.empty compare_idents) r 
	    (fallthru_cont memory lcs)
     )
  | Ycase (e, cases, opt) ->
      begin
        (* Generate code for the value e which is cased on *)
    	let ((rawe,clist) as e', c') = cogen_val env e in
    	let (rcase, spill_i, layout) = 
	  match e' with
	    (Reg r, []) ->
	      begin 
	        (* ident of first case is used as the program point marker *)
		if is_dead_case deadvars cases e then
		  (r, [], layout)
		else
		  let (spill_i,layout') = spill_reg layout r in 
		  (r, spill_i, layout')
	      end
	  | _ -> 
	      let (r, spill_i, layout') = get_reg layout in
	      (r, Mov (Reg r, e') :: spill_i, layout')
    	in
	let env = set_layout env layout in    
	let (il, cbs) =
      	match next with
	  Return -> case_aux env (rcase, c', cases, opt) Return Return cont
	| Bind x -> 
	    let l = new_label "nxt" in
	    (* variables to instantiate label.  
	       The free stack and type variables should be the same. *)
	    let lcs = 
	      Cvar rho1 :: Cvar rho2 :: 
	      (Dict.fold_dict (fun v _ lcs -> Cvar v :: lcs) delta [])
	    in
	    (* Generate context as it will be on return from each branch
	       arm.  Assume all variables have been spilled.  Need 
	       an analysis to determine which registers not to
	       spill.
	    *)
	    let regs' = new_regs () in
	    let frame' = new_frame (Array.length frame) in
	    let _ = frame'.(ret) <- frame.(ret) in
	    let memvars = memvars env in
	    let _ = List.iter (fun x ->
	      let loc = Dict.lookup ra x in
	      match loc with
		Sloc i -> 
		  divisible4 i "cogen_rhs: case"; 
		  frame'.(i/4) <- Filled_slot x
	      |	Rloc r -> regs'.(reg2int r) <- Some x
	      |	_ -> die "illegal spill location") memvars
	    in
	    let env' = set_layout env (ra,regs',frame',ret,deadvars) in
	    let codetype = cogen_codetype env' in
	    let (vks,regstate) = split_codetype codetype in
	    let memory = (regs', frame') in
	    (* Destination of case always Eax - arbitrary choice do something
	       smarter later.
	    *)
	    let r = Eax in
	    case_aux env (rcase, c', cases, opt) 
	      (Jump (r, l, lcs, memory)) (Fall (r, l, lcs, memory)) 
	      (fun (c, ((ra, regs, frame, ret, deadvars) as layout)) -> 
		let regstate = Dict.insert regstate r (cogen_con c) in
		regs.(reg2int r) <- Some x;
		let (il, cbs) = cont (c, layout) in
		([], (l, make_codetype vks regstate, il) :: cbs)
		    )
	| Jump (_, _, _, _) ->
	    case_aux env (rcase, c', cases, opt) next next cont
	| Fall (r,l,lcs, memory) ->
	    case_aux env (rcase, c', cases, opt) (Jump (r,l,lcs,memory)) next cont
	in
	(spill_i @ il, cbs)
      end
  | Ycaseexn (e,cases,default) -> 
      raise (Cogen_unimplemented "cogen_rhs: Ycaseexn")
  | Ytry (e, x, handle) -> raise (Cogen_unimplemented "cogen_rhs: Ytry")
  | Yarray es -> raise (Cogen_unimplemented "cogen_rhs: Yarray")
  | Yvector es -> raise (Cogen_unimplemented "cogen_rhs: Yvector")

  (* Everything else is a simple computation *)
  | _ ->
      begin
	match next with
	  Return -> 
	    (* Must return after translating the next instruction.  Any
	       continuation is dead code so throw away the continuation
	    *)
	    cogen_simple_rhs env rhs (Bind_loc Eax) return_cont
	| Bind x -> 
	    cogen_simple_rhs env rhs (Bind_id x) cont
	| Jump (r,l,cs,memory) ->
	    cogen_simple_rhs env rhs (Bind_loc r) (jump_cont memory l cs)
	| Fall (r,l,cs,memory) ->
	    cogen_simple_rhs env rhs (Bind_loc r) (fallthru_cont memory cs)
      end

(* Generate Code for expressions *)
and cogen_exp 
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env : env) 
    next e cont
      : (instruction list * code_blockl list) =
 begin
  match e with
    Ylet (Ytypebind (v,c), e) ->
      let kappa' = Dict.insert kappa v (ynorm kappa c) in
      cogen_exp ((delta, gamma, kappa'), layout, taldelta) next e cont
  | Ylet (Yvalbind (x, Yunpack (rhs,a,y,e1)), e) ->
      cogen_exp env next 
	(Yunpack (rhs, a, y, (Ylet (Yvalbind (x,e1), e)))) cont
  | Ylet (Yvalbind (x1, Ylet (xrhs,e1)), e2) ->
      cogen_exp env next (Ylet (xrhs, Ylet (Yvalbind (x1, e1), e2))) cont
  | Ylet (Yvalbind (x,rhs), e) ->
      cogen_rhs env (Bind x) rhs
	(fun (c, layout)  -> 
	  let gamma' = Dict.insert gamma x c in
	  cogen_exp ((delta, gamma', kappa), layout, taldelta) next e cont)
  | Ylet (Yrecbind fs, e) ->
      (* We either have a letrec of all functions or a letrec of some other
	 values. 
      *)
      (match fs with
	[] -> die "cogen_exp: empty letrec"
      |	(_,_,Yabs (_,_,_))::rest -> 
	  die "cogen_exp letrec of functions only at top level"
      | _ -> 
	  die "rec of values unimplemented") (* cogen_rec env next fs cont *)
  | Yunpack (rhs, v, x, e) ->
      begin
      cogen_rhs env (Bind x) rhs
	(fun (c, layout) -> 
	  (match c with
	    YCexist (a, k, c') ->
	      let delta' = Dict.insert delta v k in
	      let gamma' = Dict.insert gamma x (ysubst (YCvar v) a c') in
	      let loc = get_loc layout x in
	      let (dest, source, layout', spill_i) =
		match loc with
		  Rloc r ->
		    (r, raw (Reg r), layout, [])
		| Sloc i -> 
		    let (r,spill_i,layout') = get_reg layout in
		    (r, raw (sloc2genop (Array.length frame) i), 
		     layout', spill_i)
		| _ -> die "unpack dest must be reg or stack"
	      in
	      let layout' = kill_vars (set_layout env layout') 
		  (Set.singleton compare_idents x)
	      in
	      let layout' = add_id layout' x (Rloc dest) in 
	      let (il,cbs) = 
		cogen_exp ((delta', gamma', kappa), layout', taldelta) 
		  next e cont
	      in
	      (spill_i @ Unpack (v, dest, source) :: il, cbs)
          | _ -> die "cogen_exp: unpack of non-existential type"))
      end
  | _ -> 
      (* e must be a right-hand-side *)
      cogen_rhs env next e cont
 end
(* Generate code for a function.
   (delta, gamma, kappa): type environment
   ra: register assignment
   f: the Ycalculus function
   c: the type of f
   l: the label to which the code should be bound
   cbs: continuing code_blocks

   result: code_blockl list
*)

let cogen_fun ((delta, gamma, kappa) as ctxt) ra deadvars framesz (x,l,c,f) cbs =
    match f with
      Yabs (vks, xcs, body) ->
	let cnorm = ynorm kappa c in
	let (vks', normcs) =
	  match cnorm with
	    YCabs (vks,cs,_) -> (vks, cs)
	  | _ -> die "cogen_fun: constructor not function type"
	in
	let (codetype, taldelta, retaddrc) = tal_function_type delta cnorm in
      	let delta, kappa = 
	  List.fold_left2 (fun (delta,kappa) (v, k) (v',k') -> 
	    (Dict.insert delta v k, Dict.insert kappa v (YCvar v'))) 
	      (delta,kappa) vks vks'
	in
	let regs = new_regs () in
	(* new frame is framesz plus room for arguments plus return addr *)
	let frame = new_frame (framesz + 1 + List.length normcs) in
	(* Arguments are placed on the stack, pushed from right to left. The
	   return address is on top.
	*)
      	let (gamma, layout, argsz) = 
	  List.fold_right2 (fun (xarg, _) cnorm (g, layout, argsz) ->
	    let current_argsz = ysizeof delta cnorm in
	    (Dict.insert g xarg cnorm, 
	     add_id layout xarg (Sloc argsz), 
	     argsz + current_argsz)
	  ) xcs normcs (gamma, (ra, regs, frame, -1, deadvars), 0)
      	in
	let ret = argsz/4 in
	frame.(ret) <- Anon_slot retaddrc;
	let layout : layout = 
	  match layout with
	    (ra, regs, frame, _, deadvars) -> (ra,regs,frame,ret,deadvars)
	in
	let ctxt : ctxt = (delta, gamma, kappa) in
  	let layout = 
	  kill_vars (ctxt, layout, taldelta) 
	    (get_deadvars deadvars x) 
	in
      	let env : env = (ctxt, layout, taldelta) in
	let cont = fun _ -> ([], cbs) in
      	let (il, cbs) = cogen_exp env Return body cont in
	if framesz > 0 then
	  (l,codetype, (ArithBin(Sub,Reg Esp,Immed (framesz*4))) :: il) :: cbs
	else
	  (l,codetype, il) :: cbs
    | _ -> die "cogen_fun: not function"
  
(* Generate code for a TAL module.  The expression e must be a series of
   lettype definitions, followed by a letrec containing the hoisted program 
   code, followed by an anormal form expression.
*)
let rec cogen_toplevel 
    (((delta, gamma, kappa) as ctxt,
      ((ra, regs, frame, ret, deadvars) as layout),
      ((rho1, rho2) as taldelta)) as env)
    framesizes e =
  match e with
    Ylet (Ytypebind (v,c), e) ->
      let kappa' = Dict.insert kappa v (ynorm kappa c) in
      cogen_toplevel (set_ctxt env (delta, gamma, kappa')) framesizes e
  | Ylet (Yrecbind fs, e) ->
      (* Add all typing information for each function to context.
	 Generate new label for each function.
      *)
      let (gamma', ra',ls) = List.fold_right (fun (x,c,f) (gamma',ra',ls) ->
	let l = new_label ("fn"^(string_of_int x)) in
	(Dict.insert gamma' x (ynorm kappa c), 
	 Dict.insert ra' x (Hloc l),
	 l::ls)) fs (gamma, ra, [])
      in
      let ctxt = (delta,gamma',kappa) in
      let f_cbs =
	List.fold_right2 (fun (x,c,f) l cbs ->
	  let framesz = (Hashtbl.find framesizes x).framesz in
	  cogen_fun ctxt ra' deadvars framesz (x,l,c,f) cbs
	) fs ls []
      in
      let layout = (ra', regs, frame, ret, deadvars) in
      let top_e = 
	cogen_exp (ctxt, layout, taldelta) Return e (fun x -> ([],[]))
      in
      (top_e, f_cbs)
      (*
      let fs_cont = fun _ -> ([], f_cbs) in
      let ctxt = (delta, gamma', kappa) in
      let layout = (ra', regs, frame, ret, deadvars) in
      cogen_exp (ctxt, layout, taldelta) Return e fs_cont 
      *)
  | _ -> die "module has no functions"

let cogen_module basename tal_int_fn (toplevel_framesz, deadvars, framesizes) 
    (YMstruct (c,e)) =
  let vectorize_cbs cbs =
    let vectorize_cb (l, ls, il) =
      let iv = Array.of_list il in
      (l, ls, iv)
    in
    Array.of_list (List.map vectorize_cb cbs)
  in

  (* Initial typing and register assignment info *)
  let delta = Dict.empty compare_vars in
  let gamma = Dict.empty compare_idents in
  let kappa = Dict.empty compare_vars in
  let ra = Dict.empty compare_idents in
  let regs = new_regs () in
  (* Initial Stack Info, Label Typing *)
  let frame = new_frame (toplevel_framesz+1) in
  let (tc, taldelta, retaddrc) = tal_function_type delta (YCabs ([],[],c)) in
  let _ = frame.(0) <-  Anon_slot retaddrc in
  let ret = 0 in
  (* Generate code for module *)
  let ctxt = (delta, gamma, kappa) in
  let layout = (ra, regs, frame, ret, deadvars) in
  let env = (ctxt, layout, taldelta) in
  let ((il,cbs), cbs') = cogen_toplevel env framesizes e in
  let xentry = label_of_string basename in
  let il = (ArithBin(Sub,Reg Esp,Immed (toplevel_framesz*4)))::il in
  let cbs = (xentry,tc,il):: cbs @ cbs' in
  let tal_int =
    { int_includes = [|"kml_types.tali"|];
      int_cons = [||];
      int_typedefs = [||];
      int_labs = [|(xentry,tc)|]
    } 
  in let tal_imp = 
    { imports = [|"kml_types.tali"|];
      typedefs = [||];
      exports = [|tal_int_fn|];
      con_blocks = [||];
      code_blocks = vectorize_cbs cbs;
      data_blocks = [||]
    } 
  in (tal_int, tal_imp)

(* GENERATING STUBS FOR LETREC OF NON-FUNCTIONS *)
(*
and cogen_rec ((delta, gamma, kappa, ra, 
		((retaddrc, ssz, sigma, rho1, rho2) as sinfo)) as env)
    next fs cont =
	  (* Return identifier option: Some x if e is a coercion of an
	     identifier x. None if it is not a coercion. die if e is
	     not a small value.
          *)
	  let rec get_var e =
	    match e with
	      Yvar x -> Some x
	    | Yfold (k,c1,c2,e) -> getvar e
	    | Yunfold (k,c1,c2,e) -> getvar e
	    | Yroll(c,e) -> getvar e
	    | Yunroll e -> getvar e
	    | Ytag (i, _, _) -> None
	    | Yprim (p,_,_) -> None
	    | Ypack (t,_,_,_,_) -> getvar t
	    | _ -> die "rec_aux: init_elem: get_var: e not small value"
	  in
	  (* Find the free type variables in constructor c assuming all
	     bound variables are unique 
	  *)
	  let empty = Set.empty compare_idents in
	  let single x = Set.singleton compare_idents x in
	  let delete s x = Set.delete s x in
	  let union s1 s2 = Set.union s1 s2 in
	  let rec fv c = 
	    match c with
	      YCvar x -> single x
	    | YCconst _ -> empty
	    | Ylam (v,_,c) -> 
		delete (fv c) v
	    | YCapp (c1, c2) ->
		union (fv c1) (fv c2)
	    | YCtuple cs ->
		List.fold_left (fun c s -> union (fv c) s) cs empty
	    | YCproj (_,c) ->
		fv c
	    | YCtag (i,c,_) ->
		fv c
	    | YCdecide (c, vkcs, _) ->
		union (fv c) 
		  (List.fold_left 
		     (fun (v,_,c) s -> union (delete (fv c) v) s) vkcs empty)
	    | YCrec (_,c) -> fv c
	    | YCmu (v,c) -> delete v (fv c)
	    | YCabs (vks, cs, c) ->
		let s = 
		  List.fold_left (fun c s -> union (fv c) s) (c::cs) empty
		in
		List.fold_left (fun (v,_) s -> delete s v) vks s
	    | YCexist (v,_,c) -> 
		delete (fv c) v
	    | YCprod cs ->
		List.fold_left (fun c s -> union (fv c) s) cs empty
	    | YCsum css ->
		List.fold_left (fun cs s -> 
		  List.fold_left (fun c s' -> union (fv c) s') cs s)
		  css empty
	    | YCunit | YCvoid -> empty
	  in
	  (* Construct a stub object of type c *)
	  let rec make_stub c il =
	    match c with
	      YCabs (vks, cs, c') ->
		(* Add the free variables to the abstraction *)
		let fvs = Set.elements (fv c) in
		let vks' = List.fold_right (fun v vks' -> 
		  (v, Dict.lookup delta v) :: vks') fvs vks
		in
		(* Generate the ycalculus function of the appropriate type *)
		let lstub = new_label "stub" in
		let xargs, xcs = 
		  List.fold_right (fun c (xargs, xcs) -> 
		    let x = new_ident () in
		    (Yvar x::xargs, (x,c)::xcs)) cs ([],[])
		in
		let ystub = Yabs (vks, xcs,
		      	Yapp (xstub, List.map (fun (v,k) -> v) vks, xargs))
		in
		(* Generate Tal function *)
		let delta = Dict.empty compare_vars in 
		let gamma = Dict.empty compare_idents in
		let ra = Dict.empty compare_idents in
		let talstub = cogen_fun delta gamma ra ystub c empty_cont in
		(* Apply the free type variables to the label *)
		let coerce_l = List.fold_left (fun v l ->
		  Tapp (l, YCvar v)) fvs (Raw (Addr lstub))
		in
		(* Put the stub in the heap and return the coerced label *)
		(add_heap talstub;
		 (coerce_l, il))
	    | YCexist (v,k,c') ->
		(* v hides some type. We can choose it, so choose YCunit *)
		let c' = ysubst YCunit v c' in
		let (stubv,il) = make_stub c' il in
		(Pack (stubv, Cprod [], cogen_con c), il)
	    | YCprod cs ->
		let (vs,il) = List.fold_right (fun c (vs,il) ->
		  let (v,il') = make_stub cs il in
		  (v::vs,il')) cs 
		in
		let rtemp = new_reg () in
		let rreg = match rtemp with
		  Reg r -> r
		| _ -> die "won't happen"
		in
		let List.fold (fun c (sz,sizes) -> 
		  let csz = ysizeof delta c in
		  (sz + csz, csz :: sizes)) cs (0,[])
		in
		(Raw rtemp,
		 il @
		 Malloc (rtemp, sz, cogen_cons cs) ::
		 List.fold_left (fun (offset,ils) csz ->
		   (offset+csz, Mov (Prjr , Raw r, offset) :: ils)) 
		   sizes (0, []))
	    | _ -> die "make_stub: can't make stub of that type"
	  in

	  (* Initialize an element e of a product of type c *)
	  let init_elem c e fixup_list il =
	  match get_var e with
	    None -> 
	      let (v,c) = cogen_val env e in
	      Mov (Reg (Virt x), v) :: il
	  | Some x ->
	    try
	      let regopt = Dict.lookup stubDict x in
	      (match regopt with
		None -> 
		  (* Stub needs to be created *)
		  let (r, stubil) = make_stub c in
		  stubil @ Mov (Reg (Virt x), r) :: il
	      |	Some reg ->
		  (* Stub has already been created and sits in reg *)
		  Mov (Reg (Virt x), Reg reg) :: il
	      )
	    with 
	      Dict.absent -> 
		let (v,c) = cogen_val env e in
		[Mov (Reg (Virt x), v)]
	in

  (* For each element (x,c,e) of the letrec, given:
     stubdict: maps elements of letrec to None if no stub has been created
       and maps elements of letrec to Some of reg if the stub has been
       created and stored in reg
     cont: mapping from an (xalias, offset, ident) list to continuing 
       instructions and code blocks.  The xalias must be filled in at offset
       with the letrec element bound to ident.
  *)
  	let rec rec_aux 
	    (x,c,e) 
	    (stubDict : (ident, reg option) Dict.dict) 
	    (cont : instruction list -> 
	      (instruction list * code_blockl list)) =
	  (match e with
	    Ypack (t,v,k,c1,c2) ->
	      rec_aux
	  | Yprod es ->

    | _ -> die "cogen_rec: non-product/non-pack letrec"
	  )
	in
	(* cogen_rec body *)
	List.fold_left (fun (f,c,e) (stubDict,cont) ->
	  rec_aux (f,c,e) stubDict cont) (Dict.empty compare_idents)
	  (fun il -> let (il', cb) = cont () in
                     (il @ il', cb))

*)
