(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

open Format
open Ypp
open Utilities
open Ycon
open Ycalculus
open Ident
open Variable
open Yverify

exception Analyze of string
exception Analyze_unimplemented of string

let die s =  (print_string "Cogen Analyze error: ";
	      print_string s;print_string"\n";
              raise (Analyze s))

let protect f x = 
    (Format.open_box 0; f x; Format.close_box();Format.print_flush())
let dumpc c = protect Ypp.print_con c
let dumpt t = protect Ypp.print_term t
let dumpk k = protect Ypp.print_kind k

let print_set f s = 
  let rec print_list f l =
    match l with
      [] -> ()
    | [x] -> f x
    | hd::tl -> f hd; print_string ", "; print_list f tl
  in
  let l = Set.elements s in
  protect (fun _ -> print_char '{'; print_list f l; print_char '}') ()

let print_table f t =
  Hashtbl.iter 
    (fun x i -> protect 
	(fun _ -> 
	  print_int x; print_string ": "; f i; print_cut ()) ()) 
    t

let rec fold_right3 f l1 l2 l3 base =
  match (l1,l2,l3) with
    ([],[],[]) -> base
  | (x1::xs1, x2::xs2, x3::xs3) -> 
      f x1 x2 x3 (fold_right3 f xs1 xs2 xs3 base) 
  | _ -> die "fold_right3: list length does not agree"

let rec fold_right4 f l1 l2 l3 l4 base =
  match (l1,l2,l3,l4) with
    ([],[],[],[]) -> base
  | (x1::xs1, x2::xs2, x3::xs3, x4::xs4) -> 
      f x1 x2 x3 x4 (fold_right4 f xs1 xs2 xs3 xs4 base) 
  | _ -> die "fold_right4: list length does not agree"

(* Information stored at each program point (identifier binding point) *)
type info = {deadvars : ident Set.set}
let print_info = protect (fun i -> print_set print_int i.deadvars)
let print_itable = print_table print_info

(* Information about each function *)
type fninfo = {framesz : int}
let print_fninfo = protect (fun fni -> print_int fni.framesz)
let print_fntable = print_table print_fninfo 

(* Information associated with each case statement *)
type caseinfo = {switchdead : ident Set.set; defdead : ident Set.set}
let print_caseinfo = protect (fun csei -> 
  print_set print_int csei.switchdead; 
  print_string " ** ";
  print_set print_int csei.defdead) 
let print_casetable = print_table print_caseinfo

(* Print results of Analysis *)
let print_analysis (toplevel_framesz, deadvars, framesizes) =
  print_string ("toplevel_framesz: "^(string_of_int toplevel_framesz));
  print_newline ();
  print_string ("dead_vars:\n");
  print_itable (fst deadvars);
  print_newline ();
  print_string ("case_deadvars:\n");
  print_casetable (snd deadvars);
  print_newline ();
  print_string ("frame_sizes:\n");
  print_fntable framesizes;
  print_newline ()

(* Set manipulation Abbreviations *)
let empty : ident Set.set = Set.empty compare_idents
let single i : ident Set.set = Set.singleton compare_idents i
let set_minus s l = List.fold_right (fun x s -> Set.delete s x) l s
let set_unions ss =
  List.fold_right (fun s s' -> Set.union s s') ss empty
let intersect s1 s2 = 
  List.fold_left (fun s x -> 
    if Set.member s2 x then Set.insert s x
    else s) empty (Set.elements s1)
let max a b = if a > b then a else b
let maxes l = List.fold_left max 0 l

(* Information passed up through analyze:
   max: maximum number of variables live in any subexpression
   live: the live variables
*)
type ret = {max : int; live : ident Set.set}

(* compute (framesz, (deadvars, casedead), framesizes) where
   framesz = framesize of toplevel expression
   deadvars = hastable mapping let-bound identifiers to the variables
               which become dead at that point in the code
   casedead = hastable as above. ident in first branch of case maps to
              the variables which die after the switch and die in
              the default.
   framesizes = hashtable mapping function identifiers to the size of
                frame (not including the space needed for arguments)
*)
let analyze m =
  (* the table mapping function identifiers to information *)
  let fntable : (ident,fninfo) Hashtbl.t = Hashtbl.create 31 in
  let insert_new_fn fn fninfo = Hashtbl.add fntable fn fninfo in

  (* the table mapping let-bound identifiers to live variable info *)
  let infotable : (ident, info) Hashtbl.t = Hashtbl.create 31 in
  let insert_info x info = Hashtbl.add infotable x info in

  (* table mapping the first ident in a case to the set of idents which are
     dead after switching on the initial expression and dead in the
     default.
  *)
  let casetable : (ident, caseinfo) Hashtbl.t = Hashtbl.create 31 in
  let insert_case cases caseinfo = 
    match cases with
      [] -> () (* this shouldn't happen - all cases should have a branch *)
    | (_,x,_)::rest -> Hashtbl.add casetable x caseinfo
  in
 
  (* Compute the variables that die at the binding site x *)
  (* object dies at x if it is live in rhs but not in e *)
  (* add these objects to infotable *)
  let compute_dead x rhs_live body_live =
    let deadvars =
      List.fold_left (fun s xrhs -> 
      	if Set.member body_live xrhs then s
      	else Set.insert s xrhs
	    ) empty (Set.elements (Set.insert rhs_live x))
    in
    insert_info x {deadvars=deadvars}
  in

  (* tpass: a recurses over right-hand sides and values
     id: the identifier to which the expression will be bound
     labelSet: the idents in this set are labels - not included in live var
               calculation
     livevars: the idents in this set are live at this point in the computation
  *)
  let rec tpass labelSet livevars e =
      match e with
      Yvar x ->
 	let livevars = 
	  if Set.member labelSet x then livevars 
      	  else Set.insert livevars x 
	in
	{max = Set.cardinality livevars; live = livevars}
    | Yprim (p,_,es) -> tpasses labelSet livevars es
    | Yabs (vks, xcs, e) -> die "abstraction not at toplevel"
    | Yapp (e1, cs, es) ->
	tpasses labelSet livevars (e1::es)
    | Ytyapp (e1, _) -> tpass labelSet livevars e1
    | Ypack (e,_,_,_,_) -> tpass labelSet livevars e
    | Ytuple es -> tpasses labelSet livevars es
    | Yproj (i,e) -> tpass labelSet livevars e
    | Ytag (_,_,es) -> tpasses labelSet livevars es
    | Ycase (e, cases, opt) ->
	(* Return (ret, xflag) for the case branch where
	   xflag is true if x is live in the branch and false otherwise
	*)
	let case_pass (_,x,e) =
	  let ret = tpass labelSet livevars e in
	  let live = Set.delete ret.live x in
	  ({max = ret.max; live = live}, Set.member ret.live x)
	in
	let (caseret, cases_live) = 
	  List.fold_right (fun case (ret, cases_live) ->
	    let (ret', xflag) = case_pass case in
	    ({max = max ret'.max ret.max; live = Set.union ret'.live ret.live},
	     (ret', xflag) :: cases_live)
	  ) cases ({max = 0; live = empty}, [])
	in
	let (branchesret, def_live_opt) =
	  (match opt with
	    None -> (caseret, None)
	  | Some default ->
	      let ret' = tpass labelSet livevars default in
	      let live = Set.union ret'.live caseret.live in
	      ({max = maxes [caseret.max; ret'.max; Set.cardinality live];
		live = live}, Some ret'.live))
	in 
	(* Compute the variables dead upon entry into a given branch of the
	   case *)
	let total_live = Set.elements branchesret.live in
	let _ =
	  List.map2 (fun (_,x,_) (ret, xflag) ->
	    let xdead = if xflag then empty else Set.insert empty x in
	    let xdead = List.fold_left (fun xdead y ->
	      if Set.member ret.live y then xdead
	      else Set.insert xdead y) xdead total_live
	    in
	    insert_info x {deadvars = xdead}
	    ) cases cases_live
	in
	(* Compute the variables dead upon entry to the default. TODO. 
	   Nowhere to store them.
	*)
	let def_dead = 
	  match def_live_opt with
	    None -> empty
	  | Some def_live ->
	      List.fold_left (fun def_dead y ->
	  	if Set.member def_live y then def_dead
	  	else Set.insert def_dead y) empty total_live
	in
	(* Compute the variables dead after switching on e *)
	let eret = tpass labelSet branchesret.live e in
	let edead =
	  List.fold_left (fun s x ->
	    if Set.member branchesret.live x then s
	    else Set.insert s x) empty (Set.elements eret.live)  
	in
	insert_case cases {switchdead = edead; defdead = def_dead};
	let live = Set.union eret.live branchesret.live in
	{max = maxes [eret.max; branchesret.max; Set.cardinality live]; 
	live = live}
    | Ybuildexn (e1,e2) ->
	tpasses labelSet livevars [e1;e2]
    | Ytry (e,x,handler) ->
	let ret_handler = tpass labelSet livevars handler in
	let live_handler = Set.delete ret_handler.live x in
	let ret1 = tpass labelSet livevars e in
	let live = Set.union live_handler ret1.live in
	{max = maxes [ret1.max; ret_handler.max; Set.cardinality live];
	 live = live}
    | Ycaseexn (e, cases, default) ->
	raise (Analyze_unimplemented "Ycaseexn")
	(*
	let case_pass (e1,x,e2) =
	  let ret2 = tpass labelSet livevars e2 in
	  let ret1 = 
	    tpass labelSet (Set.union (Set.delete ret2.live x) livevars) e1 in
	  compute_dead x ret1.live (Set.union ret2.live livevars); 
	  let live = Set.union ret1.live (Set.delete ret2.live x) in
	  {max = maxes [ret1.max; ret2.max; Set.cardinality live]; 
	   live = live}
	in
	let caseret = List.fold_left (fun ret case ->
	  let ret' = case_pass case in
	  {max = max ret'.max ret.max; live = Set.union ret'.live ret.live}
	) (tpass labelSet livevars default) cases
	in
	let eret =  tpass labelSet caseret.live e in
	let edead = 
	  List.fold_left (fun s xrhs -> 
      	    if Set.member caseret.live xrhs then s
      	    else Set.insert s xrhs
	    	) empty (Set.elements eret.live)
	in
	insert_case cases {deadvars = edead};
	{max = maxes [eret.max; caseret.max; Set.cardinality eret.live];
	  live = eret.live}
	*)
    | Yfold (_,_,_,e) -> tpass labelSet livevars e
    | Yunfold (_,_,_,e) -> tpass labelSet livevars e
    | Yroll (_, e) -> tpass labelSet livevars e
    | Yunroll e -> tpass labelSet livevars e
    | Yarray es -> tpasses labelSet livevars es
    | Yvector es -> tpasses labelSet livevars es
    | Ycons (e1,e2) -> 
	let ret1 = tpass labelSet livevars e1 in
	let ret2 = tpass labelSet livevars e2 in
	let live = Set.union ret1.live ret2.live in
	{max = maxes [ret1.max; ret2.max; Set.cardinality live]; live = live}
    | Ylet (Ytypebind _, e) -> tpass labelSet livevars e
    | Ylet (Yrecbind fs, e) -> raise (Analyze_unimplemented "let rec")
    | Ylet (Yvalbind (x, Yunpack (rhs,a,y,e1)), e) ->
	tpass labelSet livevars 
	  (Yunpack (rhs, a, y, (Ylet (Yvalbind (x,e1), e))))
    | Ylet (Yvalbind (x1, Ylet (xrhs,e1)), e2) ->
	tpass labelSet livevars (Ylet (xrhs, Ylet (Yvalbind (x1, e1), e2))) 
    | Ylet (Yvalbind (x,rhs), e) ->
	let ret = tpass labelSet livevars e in
	let rhsret = tpass labelSet 
	    (Set.union (Set.delete ret.live x) livevars) rhs in
	begin
	  compute_dead x rhsret.live ret.live;
	  let live = Set.union (Set.delete ret.live x) rhsret.live in
	  (* Add 1 for x which is live *)
	  {max = maxes [ret.max; rhsret.max + 1; (Set.cardinality live)]; 
	   live = live}
	end
    | Yunpack (rhs,a,x,e) -> 
	tpass labelSet livevars (Ylet (Yvalbind (x,rhs),e))

  and tpasses labelSet livevars es =
    let ret = 
      List.fold_right (fun e ret ->
      	let ret' = tpass labelSet livevars e in
      	let live = Set.union ret.live ret'.live in
      	{max = maxes [ret.max; ret'.max; Set.cardinality live] ; live = live}
      ) es {max = 0; live = empty}
    in
    let livevars = Set.union livevars ret.live in
    {max = max (Set.cardinality livevars) ret.max; live = livevars}
  in
  let rec toplevel e =
    match e with
      Ylet (Ytypebind _,e) -> toplevel e
    | Ylet (Yrecbind fs, e) ->
	let abspass labelSet (x,c,e) =
	  match e with 
	    Yabs (vks, xcs, body) ->
	      let ret = tpass labelSet empty body in
	      let dead_args =
		  List.fold_left (fun s (x,_) ->
		    if Set.member ret.live x then s
		    else Set.insert s x) empty xcs
	      in  
	      let framesz = ret.max - (List.length xcs) in
	      insert_new_fn x {framesz = max framesz 0};
	      insert_info x {deadvars = dead_args}
	  | _ -> die "toplevel let-code contains non-function"
	in
	let labelSet = List.fold_left (fun s (x,_,_) -> 
	  Set.insert s x) empty fs
	in
	List.map (abspass labelSet) fs;
	tpass labelSet empty e
    | _ -> tpass empty empty e
  in
  let ret = match m with YMstruct (_,e) -> toplevel e in
  (ret.max, (infotable, casetable), fntable)



