(**********************************************************************)
(* (c) Greg Morrisett, Neal Glew, David Walker,                       *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Code Generation Environment and Routines for Register Allocation *)

open Utilities
open Ident
open Variable
open Tal
open Talcon
open Yprim
open Ycon
open Ycalculus
open Format
open Talpp
open Cogen_error
open Cogen_analyze
open Cogen_con

(*** Abbreviations ***)

let byte4 = Cprim(PCbytes Byte4)            (* 4 byte/integer type *)
let raw a : 'a coerce = (a, [])
let coerce (a,clist) c : 'a coerce = (a,c::clist) 

(* Code block with instruction list rather than instruction vector *)
type code_blockl = label * con * instruction list

(*** Environment and Register Allocation ***)

(* Code Generation env:
   ctxt: Ycalculus Type checking information:
     delta: a dictionary mapping Variable -> Kind
     gamma: a dictionary mapping Ident -> normalized Con
     kappa: a substitution mapping Variable -> normalized Con
   
   layout: information about object locations in the tal abstract machine
     ra: a dictionary mapping Ident -> location
         - location is the spill location of the identifier
     regs: reg -> ident option (Some x => x resides in this reg)
     frame: the last element in the frame is the top of the stack.
            the 0th element in the frame is deep within the stack.
            an array of stack_slot 
              Filled_slot x => x is in that slot  
              Anon_slot c => unnamed object of type c in that slot 
                             (return address)
              Empty_slot con => the slot contains no identifier but an
                object of type con
     ret: int (offset from bottom of frame indicating location of return
       address)
     deadvars: dictionary mapping ident -> ident Set.set (after the binding
       point for the first ident, the ident Set becomes dead)

   taldelta: tal type variables
     rho1, rho2: stack type variables of standard calling convention
     (rho1) between the current stack frame and the current exception handler
     (rho2) between the current exception handler and the bottom of the stack
   
*)

type stack_slot =
    Filled_slot of ident  (* filled with object x *)
  | Anon_slot of con      (* filled with unnamed object of type con *)
  | Empty_slot of con     (* Not filled. con is either junk or variable is
			     dead *)

type location = 
    Rloc of reg
  | Hloc of label
  | Sloc of int (* offset from the beginning of the stack frame in bytes. *)

type ra_type = (ident, location) Dict.dict
type regs_type = ident option array
type frame_type = stack_slot array
type ret_type = int
type deadvars_type = ((ident, info) Hashtbl.t * (ident, caseinfo) Hashtbl.t)

type delta = (variable, ykind) Dict.dict 
type gamma = (ident, ycon) Dict.dict
type kappa = (variable, ycon) Dict.dict

type ctxt =  delta * gamma * kappa
type layout = ra_type * regs_type * frame_type * ret_type * deadvars_type
type taldelta = variable * variable

type env = ctxt * layout * taldelta

(* Utilities *)
let set_ctxt (ctxt, layout, taldelta) ctxt' = (ctxt', layout, taldelta)
let set_layout (ctxt, layout, taldelta) layout' = (ctxt, layout', taldelta)
let set_taldelta (ctxt, layout, taldelta) taldelta' = (ctxt, layout, taldelta')
let get_ctxt (ctxt, layout, taldelta) = ctxt
let get_layout (ctxt, layout, taldelta) = layout
let get_taldelta (ctxt, layout, taldelta) = taldelta

let new_regs () = Array.create 8 None
let new_frame i = Array.create i (Empty_slot (Cprim PCjunk))

let copy_regs regs = Array.copy regs
let copy_frame frame = Array.copy frame

let copy_env (ctxt, (ra,regs,frame,ret,deadvars), taldelta) =
  (ctxt, (ra, copy_regs regs, copy_frame frame, ret, deadvars), taldelta)

let vector_fold f a lower upper base =
  let rec fold_right f i base =
    if i < lower then base
    else fold_right f (i-1) (f i a.(i) base)
  in
  fold_right f upper base

let vector_foldall f a base =
  vector_fold f a 0 ((Array.length a) - 1) base

(* convert Sloc i to a genop accessing the ith slot of the frame with given
   number of slots.
*)
let sloc2genop slot_number i = Prjr (raw Esp, (slot_number*4 - 4)-i)
(* convert any loc to a genop *)
let loc2genop frame loc =
  match loc with
    Rloc r -> Reg r
  | Hloc l -> Addr l
  | Sloc i -> sloc2genop (Array.length frame) i

(* Check for divisibility by 4 *)
exception Divisible4 of string
let divisible4 i s = if (i/4)*4 != i then raise (Divisible4 s)

(* Registers *)
let reg2int r =
  match r with
    Eax -> 0
  | Ebx -> 1
  | Ecx -> 2
  | Edx -> 3
  | Esi -> 4
  | Edi -> 5
  | Ebp -> 6
  | Esp -> 7
  | Virt _ -> die "virtual registers are not numbered"

let int2reg i =
  match i with
    0 -> Eax
  | 1 -> Ebx
  | 2 -> Ecx 
  | 3 -> Edx
  | 4 -> Esi
  | 5 -> Edi
  | 6 -> Ebp
  | 7 -> Esp
  | _ -> die ("int2reg: no register numbered "^(string_of_int i))

let print_ident = Format.print_int
let dumptregs regs = 
  open_hbox (); 
  vector_foldall (fun i r () ->
    Talpp.print_reg Format.std_formatter talppopt (int2reg i);
    print_string ":";
    (match r with
      None -> print_string "-"
    | Some x -> print_ident x);
    print_string "; ") regs ();
  print_newline(); 
  close_box()

let dumptframe frame =
  open_hbox (); 
  vector_foldall (fun i slot () ->
    print_int i;
    print_string ":";
    (match slot with
      Filled_slot x -> print_ident x
    | Empty_slot c -> 
	print_string "-";
 	Talpp.print_con Format.std_formatter talppopt c; print_string "-"
    | Anon_slot c -> Talpp.print_con Format.std_formatter talppopt c);
    print_string "; ") frame ();
  print_newline(); 
  close_box()

let dumptra ra =
  open_hbox (); 
  Dict.app_dict (fun x loc ->
    print_ident x; print_string ":";
    (match loc with
      Rloc r -> Talpp.print_reg Format.std_formatter talppopt r
    | Sloc i -> print_string "s"; print_int i
    | Hloc l -> print_label Format.std_formatter l);
    print_string "; ") ra;
  print_newline(); 
  close_box()

let dump_layout (ra, regs, frame, ret, deadvars) =
  open_hbox ();
  dumptra ra;
  dumptregs regs;
  dumptframe frame;
  print_string "ret: "; print_int ret;
  print_newline ();
  close_box ()

(* generate tal stack type *)
let frame2type (env : env) =
  let (_, _, frame, _, _) = get_layout env in
  let (_, gamma, _) = get_ctxt env in
  let vector_foldleft f a lower upper base =
    let rec fold_left f i base =
      if i > upper then base
      else fold_left f (i+1) (f i a.(i) base)
    in
    fold_left f lower base
  in
  vector_foldleft (fun _ slot cstack ->
    match slot with
      Filled_slot x -> Ccons (cogen_con (Dict.lookup gamma x), cstack) 
    | Anon_slot c1 -> Ccons (c1, cstack)
    | Empty_slot c1 -> Ccons (c1, cstack)
  ) frame 0 ((Array.length frame) - 1) Cempty
 
(* generate reg_state *)
let regs2reg_state (env : env) =
  let (rho1, rho2) = get_taldelta env in
  let (_,regs,_,_,_) = get_layout env in
  let (_,gamma,_) = get_ctxt env in
  let exn_stack = exn_stack_abbrev rho2 in
  let ebp = Capp (Cprim PCstackptr, exn_stack) in
  let sigma = frame2type env in
  let esp = Capp (Cprim PCstackptr, 
		  Cappend (sigma, Cappend (Cvar rho1, exn_stack))) in
  let reg_state : register_state = 
    vector_fold (fun i slot (reg_state : register_state) ->
      match slot with
	None -> reg_state
      |	Some x -> 
	  Dict.insert reg_state (int2reg i) (Cprim PCjunk)
(* ((cogen_con (Dict.lookup gamma x)) : con) *)
    ) regs 0 5 ((Dict.empty compare_regs): register_state)
  in
  let reg_state = Dict.insert (Dict.insert reg_state Esp esp) Ebp ebp in
  reg_state

(* Return location of identifier id.  Check to see if it is cached in
   registers first before returning the location it may have been spilled to.
*)
exception Found of location
let get_loc ((ra,regs,frame,ret,deadvars) as layout) id : location =
  try
    let loc = 
      try Dict.lookup ra id with
      	Dict.Absent -> die ("get_loc: id not found: "^(string_of_int id))
    in
    (match loc with
      Hloc l -> raise (Found (loc))
    | _ -> ());
    for i = 0 to Array.length(regs)-1 do
      match regs.(i) with
      	None -> ()
      |	Some x -> if x = id then raise (Found (Rloc (int2reg i))) else ()
    done;
    (match loc with
      Sloc i -> 
	divisible4 i "get_loc";
	if frame.(i/4) = Filled_slot id then raise (Found (loc))
      	else (dump_layout layout; 
	      die ("get_loc: id not in register but hasn't been spilled"^
		  (string_of_int id))) 
    | _ -> print_int id; print_string " "; die "get_loc: must spill to stack")
  with
    Found loc -> loc

(* spill_reg: spill register r if necessary.
   return (spill instructions, layout') 
*)
let spill_reg ((ra, regs, frame, ret, deadvars) as layout : layout) r =
  let reg_num = reg2int r in
  match regs.(reg_num) with
    None ->
      (* No need to spill - no object in register *)
      ([], layout)
  | Some x ->
	  let loc =
      	    try Dict.lookup ra x with
	      Dict.Absent -> die "get_reg: id not found"
	  in
	  begin
      	    regs.(reg_num) <- None;
            match loc with
	      Sloc i ->
	    	divisible4 i "spill_reg";
	    	if frame.(i/4) = Filled_slot x then  
            (* No need to spill - object is already in its stack slot *)
    		  ([], layout)
	    	else
            (* Must spill *)
		  begin
	      	    divisible4 i "get_reg";
	    	    frame.(i/4) <- Filled_slot x;
	    	    ([Mov (sloc2genop (Array.length frame) i, raw (Reg r))],
		     (ra, regs, frame, ret, deadvars))
		  end
	    | _ -> die "spill_reg: must spill to stack"
	  end

(* clear_reg: mov value from register r to some other location.
   Either spilling to stack or moving to another register
*)
let clear_reg ((ra, regs, frame, ret, deadvars) as layout : layout) r =
  let reg_num = reg2int r in
  match regs.(reg_num) with
    None ->
      (* No need to spill - no object in register *)
      ([], layout)
  | Some x ->
      (* Try and find a free register, other than Eax, to spill to first *)
      let i = ref 5 in
      let ans = ref ([], layout) in
      while !i != 0 do
 	if regs.(!i) = None then
	  begin
	    regs.(reg_num) <- None;
	    ans := ([Mov (Reg (int2reg !i), raw (Reg (int2reg reg_num)))], 
		    layout);
	    i := 0
	  end
	else i := !i - 1
      done;
      match !ans with
	([i], layout) -> !ans
      |	_ ->
	  (* No register was found to spill to *)
	  let loc =
      	    try Dict.lookup ra x with
	      Dict.Absent -> die "get_reg: id not found"
	  in
	  begin
      	    regs.(reg_num) <- None;
            match loc with
	      Sloc i ->
	    	divisible4 i "spill_reg";
	    	if frame.(i/4) = Filled_slot x then  
            (* No need to spill - object is already in its stack slot *)
    		  ([], layout)
	    	else
            (* Must spill *)
		  begin
	      	    divisible4 i "get_reg";
	    	    frame.(i/4) <- Filled_slot x;
	    	    ([Mov (sloc2genop (Array.length frame) i, raw (Reg r))],
		     (ra, regs, frame, ret, deadvars))
		  end
	    | _ -> die "spill_reg: must spill to stack"
	  end


(* get_reg: return a free register, spilling if necessary. *)
(* Assertion: get_reg does not move live objects already on the stack *)
(* return spill instructions, layout' *)
(* Using general purpose registers 0 - 5.  Ie: not Ebp, Esp *)
exception Empty_reg of reg
let get_free_reg ((ra, regs, frame, ret, deadvars) as layout : layout) =
  try 
    for i = 5 downto 0 do
      if regs.(i) = None then raise (Empty_reg (int2reg i))
      else ()
    done;
    None
  with
    Empty_reg r -> Some r

let get_free_reg_not ((ra, regs, frame, ret, deadvars) as layout : layout) rlist =
  try 
    for i = 5 downto 0 do
      if regs.(i) = None then 
	let r = int2reg i in
	if List.mem r rlist then
	  ()
	else
	  raise (Empty_reg r)
      else ()
    done;
    None
  with
    Empty_reg r -> Some r

let last_spilled = ref 0
let pick_victim layout =
  let next_spilled = 
    if !last_spilled != 0 then
      5
    else
      !last_spilled - 1
      
  in
  last_spilled := next_spilled;
  next_spilled

let get_reg ((ra, regs, frame, ret, deadvars) as layout : layout) =
  let r = get_free_reg layout in
  match r with
    None ->
      (* If none found, spill *)
      let next_spilled = pick_victim layout in 
      let r = int2reg next_spilled in
      let (spill_i, layout') = spill_reg layout r in
      (r, spill_i, layout')
  | Some r ->
      (r, [], layout)

(* As above, except returned reg must be taken from rlist *)
let get_reg_from  ((ra, regs, frame, ret, deadvars) as layout : layout) rlist =
  match rlist with
    [] -> die "get_reg_from: empty reglist"
  | hd::tl -> 
      (* Find empty register *)
      let r = 
    	try 
	  List.fold_left (fun _ r -> 
	    let i = reg2int r in
      	    match regs.(i) with
	      None -> raise (Empty_reg r)
	    | Some x -> ()) () rlist;
	  None
    	with
	  Empty_reg r -> Some r
      in
      (* If there is no empty register, pick a victim to spill *)
      match r with
    	None -> 
	  let (spill_i, layout') = spill_reg layout hd in
	  (hd, spill_i, layout')
      |	Some r -> (r, [], layout)

(* As above, except must not disturb registers in rlist *)
let get_reg_not ((ra, regs, frame, ret, deadvars) as layout : layout) rlist =
  let r = get_free_reg_not layout rlist in
   (* If there is no empty register, pick a victim to spill *)
  match r with
    None ->
      let r = 
      	try
      	  for i = 5 downto 0 do
	    let r = int2reg i in
	    if List.mem r rlist then ()
	    else (raise (Empty_reg r))
      	  done;
	  die "rlist contains all registers!"
      	with
	  Empty_reg r -> r
      in
      let (spill_i, layout') = spill_reg layout r in
      (r, spill_i, layout')
  | Some r -> (r, [], layout)

(* check if v is dead at program point x given deadvars *)
let is_dead (deadvars : deadvars_type) x v =
  let deadvars = fst deadvars in
  match v with
    Yvar y ->
      Set.member (Hashtbl.find deadvars x).deadvars y
  | _ -> false

let get_deadvars deadvars x =
  let deadvars = fst deadvars in
  (Hashtbl.find deadvars x).deadvars

let get_deadvars_case deadvars x =
  let deadvars = snd deadvars in
  (Hashtbl.find deadvars x).switchdead

let is_dead_case (deadvars : deadvars_type) cases v =
  match cases with
    [] -> false (* this shouldn't happen ... *)
  | (_,x,_)::rest -> 
      (match v with
	Yvar y ->
	  let switchdead = snd deadvars in
	  Set.member (Hashtbl.find switchdead x).switchdead y
      |	_ -> false)

(* remove variables in set s from the environment *)
let kill_vars 
    ((delta,gamma,kappa), ((ra, regs, frame, ret, deadvars) as layout), 
     taldelta : env) s =
  let kill_reg regs x =
    for i = 0 to 7 do
      if regs.(i) = Some x then
	regs.(i) <- None
      else
	()
    done
  in
  let kill_var ra regs frame gamma x =
    kill_reg regs x;
    let loc = try Dict.lookup ra x with 
      Dict.Absent -> die ("kill_var: id not found"^(string_of_int x))
    in
    (match loc with
      Sloc i -> 
	divisible4 i "kill_vars";
	if frame.(i/4) = Filled_slot x then
	  frame.(i/4) <- Empty_slot (cogen_con (Dict.lookup gamma x))
	else
	  ()
    | _ -> die "kill_var: illegal spill location");
    Dict.delete ra x
  in
  let ra = List.fold_left (fun ra x -> 
    kill_var ra regs frame gamma x) ra (Set.elements s)
  in
  (ra, regs, frame, ret, deadvars)

(* variables either on the stack or in registers *) 
let memvars (env : env) =
  let (ra, regs, frame, ret, deadvars) = get_layout env in
  let fvs = 
    vector_foldall (fun _ slot fvs ->
      match slot with
      	Some x -> x :: fvs
      | None -> fvs) regs []
  in
  vector_foldall (fun _ slot fvs ->
    match slot with
      Filled_slot x -> if List.mem x fvs then fvs else x::fvs
    | _ -> fvs) frame fvs

(* Find a stack slot not currently being used as a spill location *)
(* Return the slot number (in words from beginning of frame) *)
exception Free_stack_slot of int
let get_free_slot ((ra, regs, frame, ret, deadvars) as layout : layout) =
  let framesz = (Array.length frame) in
  let frame' = Array.create framesz false in
  Dict.app_dict (fun x loc ->
    match loc with
      Sloc i -> frame'.(i/4) <- true
    | _ -> ()) ra;
  frame'.(ret) <- true;
  try
    for i = 0 to framesz-1 do
      if frame'.(i) then ()
      else raise (Free_stack_slot i)
    done;
    dump_layout layout; die "get_free_slot: no free stack slots"
  with
    Free_stack_slot i -> i

(* add ident x to context at loc given by loc.  
   Find new location to spill to.
*)
let add_id ((ra, regs, frame, ret, deadvars) as layout : layout) x loc =
  match loc with
    Rloc r -> 
      regs.(reg2int r) <- Some x; 
      let s = get_free_slot layout in
      (Dict.insert ra x (Sloc (wordsize*s)), regs, frame, ret, deadvars)
  | Sloc i ->
      divisible4 i "add_id";
      frame.(i/wordsize)  <- Filled_slot x;
      (Dict.insert ra x (Sloc i), regs, frame, ret, deadvars)
  | _ -> die "add_id: can only add at register or stack slot"

let add_id_on_stack ((ra, regs, frame, ret, deadvars) as layout : layout) x =
  let s = get_free_slot layout in
  frame.(s) <- Filled_slot x;
  (Sloc s, (Dict.insert ra x (Sloc (wordsize*s)), regs, frame, ret, deadvars))

(* add ident x to context, replacing yvar which must be dead *)
let replace_id ((ra, regs, frame, ret, deadvars) as layout : layout) x yvar =
  match yvar with
    Yvar y -> 
      let spill_loc = Dict.lookup ra y in 
      (match get_loc layout y with
	Rloc r -> 
	  regs.(reg2int r) <- Some x; 
	  (Dict.insert ra x spill_loc, regs, frame, ret, deadvars)
      |	Sloc i ->
	  divisible4 i "replace_id";
	  frame.(i/4) <- Filled_slot x;
	  (Dict.insert ra x spill_loc, regs, frame, ret, deadvars)
      |	_ -> die "can only add to reg or plain stack ptr projection")
  | _ -> die "can't replace val; not variable"

(* Replace all instances of yvar with ident x *)
let copyprop_id ((ra, regs, frame, ret, deadvars) as layout : layout) x yvar =
  match yvar with
    Yvar y ->
      for i = 0 to (Array.length regs) - 1 do
	if regs.(i) = Some y then regs.(i) <- Some x 
	else ()
      done;
      let spill_loc = Dict.lookup ra y in
      (match spill_loc with
	Rloc r -> ()
      |	Sloc i -> 
	  divisible4 i "replace_id";
	  if frame.(i/4) = Filled_slot y then frame.(i/4) <- Filled_slot x
	  else ()
      |	_ -> die "copyprop_id: val must be on stack or in reg");
      (Dict.insert ra x spill_loc, regs, frame, ret, deadvars)
  | _ -> die "copyprop_id: can't replace val; not variable"

(* Returns true if genop is a path to memory, false otherwise *)
let is_path genop =
    match genop with
    Immed _ -> false
  | Tag _ -> false
  | Reg _ -> false
  | Addr _ -> false
  | Prjr (_,_) -> true
  | Prjl (_,_) -> true

(* return true if not both pathes *)
let valid_binops g1 g2 = not ((is_path g1) & (is_path g2))

(* If v is a Raw genop coerce and it is not a path to memory, return 
   the raw v.  Otherwise raise the exception Notplain 
*)
exception Notplain
let get_plain (v : genop coerce) : genop =
  match v with
    (v',[]) -> 
      if is_path v' then
	raise Notplain
      else
	v'
  | _ -> raise Notplain

(* Find the offset in bytes into a sequence of values with the given types *)
let get_offset delta cs i =
  let rec aux cs i offset = 
    if i = 0 then
      offset
    else
      match cs with
	[] -> die "get_offset: not enough fields"
      |	c::cs -> aux cs (i-1) (offset + ysizeof delta c)
  in
  aux cs i 0

(* perform instruction i, the result being type c, and continue with cont *)
let singlei i (c, layout) cont =
  let (il, cb) = cont (c, layout) in
  (i :: il, cb)
