(**********************************************************************)
(* (c) Michael Hicks                                                  *)
(*     May 2000, all rights reserved.                               *)
(**********************************************************************)

(* This code is used to "compile" patch files for dynamic updates.
   The basic algorithm is:

  - parse the patch file, acquiring 
    - the name of the implementation file
    - the name of the interface code file (optional)
    - the shared variables names
  - get the implementation file
    - parse it
    - for each of the shared vars v 
      - change the definition in the parsed implementation to an 
        extern.
    - for each defined v not in the shared variables
      - change the name to new__v, both in the use and the definition.
      - note a mapping XXX -> new__XXX
  - get the interface code file, if present
    - parse it
    - for each defined stub, named stub__XXX
      - if a mapping for XXX already exists (i.e. XXX -> new__XXX),
        then add to it, resulting in (XXX -> stub__XXX, new__XXX).
      - otherwise, add mapping XXX -> stub__XXX
    - append it to the modified implementation file
*)

open Popsyntax

(*************)
(* UTILITIES *)
(*************)

(* for converting :: varnames to ? ones *)
let internal_varname s =
  let rec aux news ofs =
    try
      let idx = String.index_from s ofs ':' in
      if s.[idx+1] = ':' then
	(let prefix = String.sub s ofs (idx+1-ofs) in
	prefix.[idx-ofs] <- '?';
	aux (news ^ prefix) (idx+2))
      else
	invalid_arg ("internal_varname: "^s^" is not a legal identifier")
    with Not_found ->
      news ^ (String.sub s ofs ((String.length s) - ofs)) in
  aux "" 0

(******************)
(* THE PATCH FILE *)
(******************)

(* parser *)
let parse ts =
  let error s =
    failwith ("parse error in patch file: "^s) in
  let get_token () = Stream.next ts in
  let peek () = 
    match Stream.peek ts with 
      Some t -> t
    | None -> raise Stream.Failure in

  (* returns the name of the implementation file *)
  let get_impl () =
    match get_token () with
      "implementation:" -> get_token ()
    | s -> error ("expecting \"implementation:\", got "^s) in

  (* returns the name of the interface code file *)
  let get_interface () =
    try
      (match peek () with
	"interface:" -> 
	  get_token (); (* flush it *)
	  let filename = get_token () in Some filename
      | _ -> None (* no interface file *))
    with Stream.Failure -> None (* no interface file *) in

  (* returns the list of shared variables *)
  let get_shared_vars () =
    let rec get_vars vs =
      match
	try
	  let v = get_token () in Some v
	with Stream.Failure -> None
      with
	Some v -> get_vars (v::vs)
      |	None -> vs in
    try
      (match get_token () with
	"sharing:" -> get_vars []
      | _ -> error "expecting \"sharing\"") 
    with Stream.Failure -> [] in

  let impl_fname = get_impl () in
  let ifc_fname = get_interface () in
  let svars = get_shared_vars () in
  (impl_fname, ifc_fname, svars)

let tokenize s =
  let rec skip_whitespace s i l =
    if i = l then i
    else
      match s.[i] with 
	' ' | '\n' | '\r' | '\t' -> skip_whitespace s (i+1) l
      |	_ -> i in
  let rec get_token s ofs i l =
    if i = l then
      (String.sub s ofs (i-ofs), i)
    else
      match s.[i] with
	' ' | '\n' | '\r' | '\t' ->
	  let token = String.sub s ofs (i-ofs) in
	  (token, skip_whitespace s (i+1) l)
      | _ -> get_token s ofs (i+1) l in
  let rec get_tokens s i l =
    if i = l then []
    else
      let (token,i) = get_token s i i l in
      token::(get_tokens s i l) in
  let slen = (String.length s) in
  let i = skip_whitespace s 0 slen in
  let ts = get_tokens s i slen in
  Stream.of_list ts

let get_string_from_file filename =
  try
    let fin = open_in_bin filename in
    let len = in_channel_length fin in
    let buf = String.create len in
    let res = input fin buf 0 len in
    if res <> len then
      (let ebuf = Printf.sprintf "Error reading file %s; len=%d, ret=%d"
          filename len res in
      prerr_endline ebuf; raise Exit);
    close_in fin;
    buf
  with
    Sys_error e ->
      let ebuf = Printf.sprintf "Couldn't open file %s" filename in
      failwith ebuf
  | Invalid_argument e ->
      failwith "Bad string length on input"

(* opens the patch file and parses it, returning the implementation
   file name, the interface code filename, and the shared variables *)

let get_patch_file fname =
  (* XXX deal with failure to open the file *)
  let filestr = get_string_from_file fname in
  let ts = tokenize filestr in
  parse ts

(***************************)
(* THE IMPLEMENTATION FILE *)
(***************************)

(* All of the code below is used to translate the names of variables in the
   new implementation file to be different from the names of shared
   variables *)

let new_var v = "New?"^v

type env = { local_env: string list;  (* needed to see if globals
					 are shadowed *)
	     open_vals :              (* vars inside open statement *)
	       (string, string) Dict.dict;
	     globals:                 (* all of the value globals *)
	       (string,typ) Dict.dict;
	     vars: string list;       (* list of variables to update *)
	   }

let open_prefix env p =
  begin
    let pLen = String.length p in
    let has_prefix n = 
      if String.length n <= pLen then None
      else 
	let rec aux i = 
	  (* foo?bar foo is the prefix.
	     When opened we get bar not ?bar so we need to add 1. *)
	  if i=pLen then Some (String.sub n (i+1) (String.length n - pLen - 1))
	  else if n.[i] <> p.[i] then None else aux (i+1)
	in
	aux 0
    in
    let process_dict opened d =
      let process_elt id b opened =
	match has_prefix id with
	  None -> opened
	| Some id' -> (Dict.insert opened id' id)
      in
      Dict.fold_dict process_elt d opened
    in
    let process_opened opened od =
      let process_elt id id' opened =
	(* id' is the fully-qualified name of this thing.  It just happens to
	   have been opened twice. *)
	match has_prefix id with
	  None -> opened
	| Some id'' -> (Dict.insert opened id'' id')
      in
      Dict.fold_dict process_elt od opened
    in
    let open_vals = process_opened env.open_vals env.open_vals in
    let open_vals = process_dict open_vals env.globals in
(*
    Printf.printf "open_prefix %s: open_vals are:\n" p;
    Dict.app_dict (fun k v -> Printf.printf "%s -> %s\n" k v)  open_vals;
    print_endline "";
*)
    { env with open_vals = open_vals }
  end

(* goes through the toplevel declarations and notes all of the global
   value variables, adding them to a dictionary. *)
let rec add_global globals (d,loc) =
  match d with
    ExceptionDecl(v,s,t) ->
      if Dict.member globals v then globals
      else Dict.insert globals v (ExnconType t)
  | FunDecl fd ->
      let v = fd.fn_name in
      let arg_typs = List.map snd fd.fn_args in
      let t = FnType(fd.fn_tyvars,fd.fn_ret_type,arg_typs) in
      if Dict.member globals v then globals
      else Dict.insert globals v t
  | ExternVal (v,t) -> 
      if Dict.member globals v then globals
      else Dict.insert globals v t
  | GlobalDecl (s,v,t,eor) ->
      if Dict.member globals v then globals
      else Dict.insert globals v t
  | PrefixDecl _ -> failwith "add_global: prefixes should have been eliminated"
  | OpenDecl(prefix,ds') -> add_globals globals ds'
  | _ -> globals
and add_globals globals ds = List.fold_left add_global globals ds

exception Unbound

(* for completing the names of variables within an open declaration *)
let complete_val_name env n =
  let defined n = List.mem n env.local_env in
  try
    if defined n then n 
    else Dict.lookup env.open_vals n
  with Dict.Absent -> n

(* indicates the binding of a local variable shadows a global one *)
let bind env v =
  { env with local_env = v::env.local_env }
let binds env vs =
  List.fold_left (fun env v -> bind env v) env vs

let dyntrans_var env v =
  let v = complete_val_name env v in
  if List.mem v env.vars &&
     not (List.mem v env.local_env) then
    new_var v
  else
    v

(* don't translate types --- wouldn't be sound *)
let rec dyntrans_typ env t = t
(*
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typs' = dyntrans_typs env in
  let dyntrans_typopt' = dyntrans_typopt env in
  match t with
    Evar (vc,tor) -> Evar (vc,ref (dyntrans_typopt' !tor))
  | ArrayType (t,eopt) -> ArrayType (dyntrans_typ' t, dyntrans_expopt' eopt)
  | FnType (vs,t,tl) -> FnType (vs,dyntrans_typ' t,dyntrans_typs' tl)
  | TupleType (c,tl) -> TupleType (c,dyntrans_typs' tl)
  | NamedType (nr, tl) -> NamedType (ref (dyntrans_var env !nr), 
				     dyntrans_typs' tl)
  | MutableTyp tr -> MutableTyp (ref (dyntrans_typ' !tr))
  | UnresolvedTyId (n,tl) -> 
      UnresolvedTyId (dyntrans_var env n, dyntrans_typs' tl)
  | RepType t -> RepType (dyntrans_typ' t)
  | ExnconType t -> ExnconType (dyntrans_typ' t)
  | _ -> t
*)

and dyntrans_typs env ts = List.map (dyntrans_typ env) ts
and dyntrans_typopt env topt =
  match topt with None -> None | Some t -> Some (dyntrans_typ env t)
and dyntrans_typsopt env tlo =
  match tlo with None -> None | Some ts -> Some (dyntrans_typs env ts)

and dyntrans_exp env exp =
  let dyntrans_exp' = dyntrans_exp env in
  let dyntrans_exps' = dyntrans_exps env in
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_stmt' = dyntrans_stmt env in
  let dyntrans_typ' = dyntrans_typ env in
  let dyntrans_typs' = dyntrans_typs env in
  let dyntrans_typopt' = dyntrans_typopt env in
  let dyntrans_typsopt' = dyntrans_typsopt env in
  let e' =
    match exp.raw_exp with
      ConstArray (el, topt) -> 
	ConstArray (dyntrans_exps' el, dyntrans_typopt' topt)
    | Var v -> Var (dyntrans_var env v)
    | Primop (op, el) -> Primop (op, dyntrans_exps' el)
    | Conditional (be,ife,elsee) -> 
	Conditional (dyntrans_exp' be, dyntrans_exp' ife, dyntrans_exp' elsee)
    | AssignOp (lhs,opopt,rhs) ->
	AssignOp (dyntrans_exp' lhs,opopt,dyntrans_exp' rhs)
    | FunCall (fexp, tlor, args) ->
	FunCall (dyntrans_exp' fexp, ref (dyntrans_typsopt' !tlor), 
		 dyntrans_exps' args)
    | TypInst (e,ts) -> TypInst (dyntrans_exp' e, dyntrans_typs' ts)
    | NewStruct (n, tlor ,fnoptelist) ->
	let dyntrans_field (fopt, e) = (fopt, dyntrans_exp' e) in
	(* Could be an exception because of parsing, so we have to
	   check *)
	NewStruct (dyntrans_var env n, 
		   ref (dyntrans_typsopt' !tlor), 
		   List.map dyntrans_field fnoptelist)
    | StructMember (e',n) -> StructMember (dyntrans_exp' e',n)
    | NewUnion (n, tlor, nf, eopt) ->
	NewUnion (n, 
		  ref (dyntrans_typsopt' !tlor), 
		  nf, dyntrans_expopt' eopt)
    | UnionMember (e',n) -> UnionMember (dyntrans_exp' e',n)
    | NewTuple el -> NewTuple (dyntrans_exps' el)
    | TupleMember (e',x) -> TupleMember (dyntrans_exp' e',x)
    | NewAbstype (n, tlor, tlor2, e) ->
	NewAbstype (n, 
		    ref (dyntrans_typsopt' !tlor), 
		    ref (dyntrans_typsopt' !tlor2), 
		    dyntrans_exp' e)
    | Subscript (e',ie) -> Subscript (dyntrans_exp' e', dyntrans_exp' ie)
    | Codegen fd -> Codegen (dyntrans_fundecl env fd)
    | Fill e -> Fill (dyntrans_exp' e)
    | NewExn (v,eopt) -> NewExn (dyntrans_var env v, dyntrans_expopt' eopt)
    | Raise e -> Raise (dyntrans_exp' e)
    | SeqExp el -> SeqExp (dyntrans_exps' el)
    | Cast (t,e') -> Cast (t,dyntrans_exp' e') (* don't bother -- must be
						  base type anyway *)
    | Fun fd -> Fun (dyntrans_fundecl env fd)
    | _ -> exp.raw_exp
  in
  { exp with raw_exp = e' }

and dyntrans_exps env el = List.map (dyntrans_exp env) el
and dyntrans_expopt env eopt =
  match eopt with None -> None | Some e -> Some (dyntrans_exp env e)

(* STATEMENTS *)

and dyntrans_stmt env stmt =
  let dyntrans_exp' = dyntrans_exp env in
  let dyntrans_exps' = dyntrans_exps env in
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_stmt' = dyntrans_stmt env in
  let dyntrans_stmtopt' = dyntrans_stmtopt env in
  let dyntrans_typopt' = dyntrans_typopt env in
  let dyntrans_arm (x,s) = (x,dyntrans_stmt' s) in
  let dyntrans_sarm sa = 
    (* figure out all of the variables bound in the context of the arm *)
    let arm_vars p =
      let do_prim_pat p =
	match p with
	  Var_pat (v,_) -> [v]
	| _ -> [] in
      match p with
      	No_pat -> []
      |	Prim_pat p -> do_prim_pat p
      |	Tuple_pat ps ->
	  List.flatten (List.map do_prim_pat ps) in
    let env' =
      List.fold_left (fun env' v -> bind env' v) env (arm_vars sa.arm_pat) in
    (* figure out if the field name is exception name to change *)
    let v' = dyntrans_var env' sa.arm_field in
    { sa with arm_field = v'; arm_body = dyntrans_stmt env' sa.arm_body } in

  let s' = 
    match stmt.raw_stmt with
      Exp e -> Exp (dyntrans_exp' e)
    | Seq (s,s2) -> Seq (dyntrans_stmt' s, dyntrans_stmt' s2)
    | Return eopt -> Return (dyntrans_expopt' eopt)
    | IfThenElse (e,s1,s2) ->
	IfThenElse (dyntrans_exp' e,dyntrans_stmt' s1,dyntrans_stmt' s2)
    | While (e,s) -> While (dyntrans_exp' e,dyntrans_stmt' s)
    | For (e1,e2,e3,s) ->
	For (dyntrans_exp' e1,dyntrans_exp' e2,dyntrans_exp' e3, 
	     dyntrans_stmt' s)
    | IntSwitch (e, isl, s) ->
	IntSwitch (dyntrans_exp' e, List.map dyntrans_arm isl, 
		   dyntrans_stmt' s)
    | CharSwitch (e,csl,s) ->
	CharSwitch (dyntrans_exp' e, List.map dyntrans_arm csl, 
		    dyntrans_stmt' s)
    | UnionSwitch (e,sal,sopt) ->
	UnionSwitch (dyntrans_exp' e, List.map dyntrans_sarm sal, 
		     dyntrans_stmtopt' sopt)
    | ExnSwitch (e,sal,sopt) ->
	ExnSwitch (dyntrans_exp' e,List.map dyntrans_sarm sal,
		   dyntrans_stmtopt' sopt)
    | Decl (v,t,eor,s) -> Decl (v,t,ref (dyntrans_expopt' !eor),
				dyntrans_stmt (bind env v) s)
    | Label (v,s) -> Label (v,dyntrans_stmt' s)
    | Do (s,e) -> Do (dyntrans_stmt' s,dyntrans_exp' e)
    | TryHandle (s1,v,s2) -> TryHandle (dyntrans_stmt' s1,v,
					dyntrans_stmt (bind env v) s2)
    | TryCatchFinally (s,sal,defopt,finopt) ->
	TryCatchFinally (dyntrans_stmt' s,List.map dyntrans_sarm sal,
			 dyntrans_stmtopt' defopt,dyntrans_stmtopt' finopt)
    | Cut s -> Cut (dyntrans_stmt' s)
    | Splice s -> Splice (dyntrans_stmt' s)
    | With (v,tor,vs,e,s) -> 
	With (v,ref (dyntrans_typopt' !tor),
	      vs, dyntrans_exp' e, dyntrans_stmt (bind env v) s)
    | _ -> stmt.raw_stmt
  in
  { stmt with raw_stmt = s' }
  
and dyntrans_stmtopt env eopt =
  match eopt with None -> None | Some e -> Some (dyntrans_stmt env e)

(* DECLARATIONS *)

and dyntrans_topdecl env ((rawdecl,loc) as decl) =
  let dyntrans_expopt' = dyntrans_expopt env in
  let dyntrans_typ' = dyntrans_typ env in
  let d' =
    match rawdecl with
      FunDecl fd -> FunDecl (dyntrans_fundecl env fd)
    | StructDecl sd ->
	let env' = env in
	  (* List.fold_left (fun env v -> (bind env v)) env sd.st_tyvars in *)
	let dyntrans_sfield (fn,cap,t) = (fn,cap,dyntrans_typ env' t) in
	StructDecl 
	  { sd with 
	    st_fields = List.map dyntrans_sfield sd.st_fields;
	    st_name = sd.st_name }
    | UnionDecl ud ->
	  let env' = env in
	  (* List.fold_left (fun env v -> (bind env v)) env ud.un_tyvars in *)
	  let dyntrans_sfield (fn,t) = (fn,dyntrans_typ env' t) in
	  UnionDecl
	    { ud with 
	      un_fields = List.map dyntrans_sfield ud.un_fields;
	      un_name = ud.un_name }
    | AbsDecl ad -> 
	let env' = env in
(*
	  List.fold_left (fun env v -> (bind env v))
	    env (ad.abs_all_tyvars@ad.abs_exist_tyvars) in
*)
	AbsDecl { ad with 
		  abs_defn = dyntrans_typ env' ad.abs_defn;
		  abs_name = ad.abs_name }
    | ExceptionDecl (v,s,t) -> 
	ExceptionDecl (dyntrans_var env v,s,dyntrans_typ' t)
    | GlobalDecl (s,v,t,eor) ->
	GlobalDecl (s,dyntrans_var env v,
		    dyntrans_typ' t, ref (dyntrans_expopt' !eor))
    | PrefixDecl (v,tds) -> 
	failwith "dyntrans_topdecl: prefixes should have been eliminated"
(*
	PrefixDecl (v, List.map (dyntrans_topdecl env) tds)
*)
    | OpenDecl (v,tds) -> 
	OpenDecl (v, List.map (dyntrans_topdecl (open_prefix env v)) tds)
    | _ -> rawdecl (* externs stay the same *) in
  (d',loc)

and dyntrans_fundecl env fd =
  let env' = env in
  (* List.fold_left (fun env v -> (bind env v)) env fd.fn_tyvars in *)
  let env'' =
    List.fold_left (fun env (v,_) -> (bind env v)) env fd.fn_args in
  let body' = (dyntrans_stmt env'' fd.fn_body) in
  let args' = List.map (function (v,t) -> (v,dyntrans_typ env t)) fd.fn_args in
  let ret_type' = dyntrans_typ env fd.fn_ret_type in
  { fd with fn_name = dyntrans_var env fd.fn_name;
            fn_body = body'; 
            fn_args = args'; 
            fn_ret_type = ret_type' }

(* For each topdecl, if its var is in the shared variable list, then
   its definition is made extern.  Otherwise, the old defintion is
   used but its name is changed to the new implementation name and is
   noted for later translation of the expression parts. *)

let translate_impl_decls decls svars =
  let rec translate_def ((rawdecl,loc) as decl) =
    let do_trans v nv s s1 =
      if (List.mem v svars) then
	if s = Abstract then
	  failwith 
	    (Printf.sprintf "shared type definition %s is abstract" v)
	else
	  ([],[s1,loc])
      else if s <> Extern then 
	([nv,v],[decl]) 
      else 
	([],[s1,loc])
    in
    match rawdecl with
      FunDecl fd ->
	do_trans fd.fn_name (new_var fd.fn_name)
	  (if fd.fn_static then Public else Static)
	  (ExternVal (fd.fn_name, FnType (fd.fn_tyvars, fd.fn_ret_type,
					  (List.map snd fd.fn_args))))
    | StructDecl sd ->
	([],snd(do_trans sd.st_name sd.st_name sd.st_scope 
		  (StructDecl { sd with st_scope = Extern })))
    | UnionDecl ud ->
	([],snd(do_trans ud.un_name ud.un_name ud.un_scope 
		  (UnionDecl { ud with un_scope = Extern })))
    | AbsDecl ad -> 
	([],snd(do_trans ad.abs_name ad.abs_name ad.abs_scope 
		  (AbsDecl { ad with abs_scope = Extern })))
    | ExceptionDecl (v,s,t) ->
	do_trans v (new_var v) s
	  (ExceptionDecl (v,Extern,t))
    | GlobalDecl (s,v,t,eor) ->
	do_trans v (new_var v) s 
	  (GlobalDecl (Extern,v,t,eor))
    | PrefixDecl (v,tds) -> 
	failwith "translate_def: prefixes should have been eliminated"
(*
	let (vars,decls) = translate_defs tds in
	(vars, [PrefixDecl (v, decls),loc])
*)
    | OpenDecl (v,tds) -> 
	let (vars,decls) = translate_defs tds in
	(vars, [OpenDecl (v, decls),loc])
    | _ -> ([],[decl])

  and translate_defs decls =
    List.fold_right
      (fun decl (vars,decls) -> 
	let (vs,ds) = translate_def decl in
	(vs@vars,ds@decls))
      decls ([],[]) in

  (* Note all variables defined by this file that are not in the
     shared variable list, and make defintions extern for those that
     are *)

  let (varmap,decls) = translate_defs decls in
  let (_,vars) = List.split varmap in
(*
  print_endline "variable list is:";
  List.iter print_endline vars;
*)
  (* For all non-shared variables, change their names so as not to
     clash with the old implementation. *)

  let env = { local_env = []; open_vals = Dict.empty compare; 
	      globals = add_globals (Dict.empty compare) decls;
	      vars = vars } in
  let decls = List.map (dyntrans_topdecl env) decls in

  (varmap,decls)

let stub_var v = "Stub?"^v

let is_stub v =
  try
    (String.sub v 0 5) = "Stub?"
  with e -> false

let stub_root v =
  if is_stub v then
    String.sub v 5 ((String.length v) - 5)
  else
    failwith 
      (Printf.sprintf 
	 "internal_error: called stub_root with non-stub |%s|" v)

(* XXX consider making all non-stub defs in ifc code file static *)
let note_stubs ifc_decls impl_varmap =

  (* adds mapping to the variable.  If mapping already present, adds
     to the domain; otherwise adds a new mapping *)
  let add_mapping varmap k v = (v,k)::varmap in
  let remove_mapping varmap k =
    let (mapping,varmap') =
      List.fold_right
	(fun (v',k') (mapping,varmap) ->
	  if mapping <> None then (mapping,(v',k')::varmap)
	  else
	    if k = k' then (Some (v',k'),varmap)
	    else (None, (v',k')::varmap))
	varmap (None,[]) in
    match mapping with
      None -> failwith 
	  (Printf.sprintf "could not remove mapping for %s from upd_varmap"
	     k)
    | Some (v,k) -> (v,k,varmap') in

  (* if this is a stub declaration, add the mapping the variable map *)
  let rec proc_stub_decls upd_varmap add_varmap tds =
    match tds with
      [] -> (upd_varmap, add_varmap)
    | (rawdecl,_)::rest ->
	let (upd_varmap, add_varmap) = 
	  (match rawdecl with
	    FunDecl { fn_name = n } ->
	      if (is_stub n) then
		(* move existing mapping from upd_varmap to
		   add_varmap *)
		let root_name = (stub_root n) in
		let (v,k,upd_varmap) = remove_mapping upd_varmap root_name in
		(add_mapping upd_varmap root_name n,
		 add_mapping add_varmap k v)
	      else
		(upd_varmap, add_varmap)
	  | PrefixDecl (v,tds) -> 
	      failwith "proc_stub_decls: prefix found---should be none"
	  | OpenDecl (v,tds) ->
	      proc_stub_decls upd_varmap add_varmap tds
	  | _ -> (upd_varmap, add_varmap)) in
	proc_stub_decls upd_varmap add_varmap rest in

  (* note all of the stub names in the variable map *)
  proc_stub_decls impl_varmap [] ifc_decls

(* entry point *)

let patchtrans impl_decls ifc_decls svars =

  (* Convert patchfile shared variable names to internal names *)
  let svars = List.map internal_varname svars in
(*
  print_endline "shared variable list is:";
  List.iter print_endline svars;
*)
  (* Eliminate any prefixes in the code files *)
  let impl_decls = Poptype.eliminate_prefixes impl_decls in
  let ifc_decls = Poptype.eliminate_prefixes ifc_decls in

  (* Now perform the translation *)
  let (impl_varmap, impl_decls) = translate_impl_decls impl_decls svars in
  let (upd_varmap, add_varmap) = note_stubs ifc_decls impl_varmap in
(*
  print_endline "final add variable map is:";
  List.iter (function (k,v) -> Printf.printf "%s -> %s\n" k v) add_varmap;

  print_endline "final update variable map is:";
  List.iter (function (k,v) -> Printf.printf "%s -> %s\n" k v) upd_varmap;
*)
  (add_varmap,upd_varmap,impl_decls@ifc_decls)

