(**********************************************************************)
(* (c) Neal Glew, David Walker                                        *)
(*     June 1998, all rights reserved.                                *)
(**********************************************************************)

(* Numtypes - provides a number of machine numeric types and ops on them.
 *)

(* Warning: Floating point support requires a custom link to the file float.c *)

type int8 = int;;
type int16 = int;;
type int32 = int;;

let i32_0 = 0;;
let i32_1 = 1;;
let i32_2 = 2;;
let i32_3 = 3;;
let i32_4 = 4;;
let i32_8 = 8;;
let i32_16 = 16;;
let i32_32 = 32;;
let i32_255 = 255;;

let int_to_int32 i = i;;
let int32_to_int i = i;;
let int32_of_string s = int_of_string s;;

let string_of_int8 i = string_of_int i;;
let string_of_int16 i = string_of_int i;;
let string_of_int32 i = string_of_int i;;

let int16_to_int32 i = i;;
let int32_to_int8 i = i;;
let int32_to_int16 i = i;;

let int8_to_byte i = Char.chr (i land 255);;
let int32_to_1byte i = Char.chr (i land 255);;
let int32_to_2bytes i = (Char.chr (i land 255),Char.chr ((i lsr 8) land 255));;
let int32_to_4bytes i =
  let b4 =
    if (i < 0) then
      ((i asr 24) land 0xFF) lor 0x80
    else
      (i lsr 24) land 0xFF in
  (Char.chr (i land 255),
   Char.chr ((i lsr 8) land 255),
   Char.chr ((i lsr 16) land 255),
   Char.chr b4)
;;

let (~-$) = (~-);;
let (+$) = (+);;
let (-$) = (-);;
let ( *$ ) = ( * );;
let (/$) = (/);;
let mod32 = (mod);;
let (=$) = (=);;
let (<>$) = (<>);;
let (<$) = (<);;
let (<=$) = (<=);;
let (>$) = (>);;
let (>=$) = (>=);;
let lnot32 i = lnot i;;
let land32 i1 i2 = i1 land i2;;
let lor32 i1 i2 = i1 lor i2;;
let lxor32 i1 i2 = i1 lxor i2;;
let lsl32 i1 i2 = i1 lsl i2;;
let lsr32 i1 i2 = i1 lsr i2;;
let asr32 i1 i2 = i1 asr i2;;

let abs32 i = abs i
let min32 i j = min i j
let max32 i j = max i j
let max_int32 = max_int
let min_int32 = min_int
let pow32 i j =
  if j <$ 0 then invalid_arg "pow32: in32 exponent must not be negative"
  else
    let result = ref 1 in
    for k = 1 to j do
      result := !result *$ i
    done;
    !result

(* integers ordered: 0 < 1 < .... < maxint < minint < ... < -1 *)
let unsigned_lt32 i j = 
  if i >=$ 0 then
    if j >=$ 0 then i <$ j
    else true
  else (* i <$ 0 *)
    if j <$ 0 then i <$ j
    else false
      
let unsigned_lte32 i j =
  if i >=$ 0 then
    if j >=$ 0 then i <=$ j
    else true
  else (* i <$ 0 *)
    if j <$ 0 then i <=$ j
    else false

(* operators with overflow checks. Raise Invalid_arg on overflow *)
let add32_over i j =
  let res = i +$ j in 
  if i >$ i32_0 && j >$ i32_0 then
    if res <=$ i32_0 then invalid_arg "add32 overflow"
    else res
  else if i <$ i32_0 && j <$ i32_0 then
    if res >=$ i32_0 then invalid_arg "add32 overflow"
    else res
  else
    res

let sub32_over i j = 
  try
    add32_over i (0 -$ j)
  with Invalid_argument _ -> invalid_arg "sub32 overflow"

let square_root = int_to_int32 32768 (* 2^15 *)
let two_pow_14 = int_to_int32 16384
let two_pow_29 = square_root *$ two_pow_14
let signed_mul32_over i j =
  let res = i *$ j in
  let i',j' = abs32 i,abs32 j in
  if i' <$ square_root && j' <$ square_root then res
  else
    let i',j' = (min32 i' j', max32 i' j') in
    if i' <=$ i32_4 && j' <=$ two_pow_29 then res
    else if i' <=$ i32_1 && j' <=$ max_int32 then res
    else invalid_arg "signed mul32 overflow"

(* laziness: unsigned multiply has the same range as signed multiply *)
let unsigned_mul32_over i j =
  if i >$ i32_0 && j >$ i32_0 then 
    try signed_mul32_over i j
    with Invalid_argument "signed mul32 overflow" -> 
      invalid_arg "unsigned mul32 overflow"
  else invalid_arg "unsigned mul32 overflow"


(* floating point numbers:
 * conversions between 3 formats are supported:
 * 1. decformat: string with format [+|-]digits.digits[(E|e)[+|-]digits] 
 *      - decimal digits; exponent is base 10
 * 2. bytes: int list where ints are in range 0..255
 *      - See Intel Developer's Manual, Volume 1, sections 7-2, 7-4
 *      - each int is 8 binary digits
 *      - float32: 1bit sign, 8bit exp, 23bit precision, exp bias = 127
 *      - float64: 1bit sign, 11bit exp, 52bit precision, exp bias = 1023 
 *      - the exponent bias is added to each binary exponent (hence negative
 *        exponents will be represented as positive numbers)
 * 3. hexformat: string of hexidecimal digits representing bytes 
 *)
type float32 = int list (* representation: bytes *)
type float64 = int list (* representation: bytes *)

let hex_to_int c =
  match c with
    '0' -> 0
  | '1' -> 1
  | '2' -> 2
  | '3' -> 3
  | '4' -> 4
  | '5' -> 5
  | '6' -> 6
  | '7' -> 7
  | '8' -> 8
  | '9' -> 9
  | 'A' | 'a' -> 10
  | 'B' | 'b' -> 11
  | 'C' | 'c' -> 12
  | 'D' | 'd' -> 13
  | 'E' | 'e' -> 14
  | 'F' | 'f' -> 15
  | _ -> invalid_arg "hex_to_int: not hex digit"

let rec hex_to_ints s i l =
  if i < l then
    let d1 = hex_to_int s.[i]   in
    let d2 = hex_to_int s.[i+1] in
    ((d1 lsl 4) lor d2)::hex_to_ints s (i+2) l
  else
    []

let hexformat_to_float32 s = 
  if String.length s = 8 then hex_to_ints s 0 8
  else invalid_arg "hexformat_to_float32: wrong size"
  
let hexformat_to_float64 s =
  if String.length s = 16 then hex_to_ints s 0 16
  else invalid_arg "hexformat_to_float64: wrong size"

let int_to_hex i =
  match i with
    0 -> '0'
  | 1 -> '1'
  | 2 -> '2'
  | 3 -> '3'
  | 4 -> '4'
  | 5 -> '5'
  | 6 -> '6'
  | 7 -> '7'
  | 8 -> '8'
  | 9 -> '9'
  | 10 -> 'A'
  | 11 -> 'B'
  | 12 -> 'C'
  | 13 -> 'D'
  | 14 -> 'E'
  | 15 -> 'F'
  | _ -> invalid_arg "int_to_hex: bad digit"

let ints_to_hexs is =
  let l = List.length is in
  let s = String.create (2*l) in
  let rec aux is index =
    match is with
      [] -> []
    | i::is -> 
      	let h1 = int_to_hex ((i lsr 4) land 0xF) in
      	let h2 = int_to_hex (i land 0xF)         in
      	s.[index]   <- h1;
	s.[index+1] <- h2;
	aux is (index+2) in
  aux is 0;
  s

let float32_to_hexformat f32 = ints_to_hexs f32
let float64_to_hexformat f64 = ints_to_hexs f64

(* bytes = lists of integers in the range 0..255 *)
let bytes_to_float32 bytes =
  if List.length bytes = 4 then bytes
  else invalid_arg "bytes_to_float32: wrong number of bytes"

let bytes_to_float64 bytes =
  if List.length bytes = 8 then bytes
  else invalid_arg "bytes_to_float64: wrong number of bytes"

let float32_to_bytes f32 = f32
let float64_to_bytes f64 = f64

(* convert a string containing the machine representation of a float to the
 * float32/float64 representation *)

let c_to_f32 s =
  let result = ref [] in
  for i = 3 downto 0 do
    result := (Char.code (s.[i])) :: !result 
  done;
  !result

let c_to_f64 s =
  let result = ref [] in
  for i = 7 downto 0 do
    result := (Char.code (s.[i])) :: !result 
  done;
  !result

let f32_to_c f32 =
  let result = String.create 4 in
  let rec aux f32 i =
    if i < 4 then
      match f32 with
	[] -> invalid_arg "f32_to_c: float32 ill-formed"
      |	hd::rest -> result.[i] <- (Char.chr hd); aux rest (i+1)
    else
      () in
  aux f32 0;
  result

let f64_to_c f64 =
  let result = String.create 8 in
  let rec aux f64 i =
    if i < 8 then
      match f64 with
	[] -> invalid_arg "f64_to_c: float64 ill-formed"
      |	hd::rest -> result.[i] <- (Char.chr hd); aux rest (i+1)
    else
      () in
  aux f64 0;
  result

(* convert decimal strings to binary floating point *)
external d_to_h32 : string -> string = "d_to_h32"
external d_to_h64 : string -> string = "d_to_h64"

let decformat_to_float32 s = c_to_f32 (d_to_h32 s)
let decformat_to_float64 s = c_to_f64 (d_to_h64 s)

let zero_f32 = 0::0::0::0::[]
let zero_f64 = 0::0::0::0::0::0::0::0::[]
let one_f32  = 63::128::0::0::[]              (* 0011 1111 1000 0000 *)
let one_f64  = 63::240::0::0::0::0::0::0::[]  (* 0011 1111 1111 0000 0000 0000 0000 0000 *)

let is_zero_f32 f32 = f32 = zero_f32
let is_zero_f64 f64 = f64 = zero_f64
let is_one_f32  f32 = f32 = one_f32 
let is_one_f64  f64 = f64 = one_f64  

(* Floating Point Negation *)
external negate_f32 : string -> string = "negate_f32"
external negate_f64 : string -> string = "negate_f32"
 
let negf32 f32 = c_to_f32 (negate_f32 (f32_to_c f32))
let negf64 f64 = c_to_f64 (negate_f64 (f64_to_c f64))
(* EOF: numtypes.ml *)



